(function() {

    window.adl = new ADL();

    window.addEventListener("load",initADL);
    window.addEventListener("keydown",onKeyDown);
    window.addEventListener("focus",setupDynamicStyles);

    const ADL_VERSION = "1.1.0";
    const IONIC_VERSION = "5.2.3";

    const DEFAULT_COLORPICKER_OPTIONS = [
        [ "#ef9a9a","#FFCC80","#FFF59D","#A5D6A7","#90CAF9","#B39DDB","#F48FB1","#BCAAA4","#ffffff" ],
        [ "#f44336","#FF9800","#FFEB3B","#4CAF50","#2196F3","#9C27B0","#E91E63","#795548","#9E9E9E" ],
        [ "#b71c1c","#E65100","#F9A825","#1B5E20","#0D47A1","#4A148C","#880E4F","#3E2723","#000000" ]
    ];

    let adlTheme = "#1565C0";

    let adlBlocker = null;
    let forceTheme = null;
    let adlDynamicStyle = null;
    let currentADLDialog = null;
    let pendingInputText = "";
    let pendingSliderValue = 0;
    let pendingColorChoice = null;
    let deviceSpecial = null;
    let toastDismissTimeout = null;
    let overallThemeLight = true;

    let adlIsInit = false;
    let toastQueue = [];

    function ADL() {

        this.setTheme = setTheme;
        this.showDialog = showDialog;
        this.showAlert = showAlert;
        this.showAlertWithTitle = showAlertWithTitle;
        this.showChoice = showChoice;
        this.inputBox = inputBox;
        this.pickColor = pickColor;
        this.showProgress = showProgress;
        this.updateProgress = updateProgress;
        this.dismissProgress = dismissProgress;
        this.showList = showList;
        this.addToolbar = addToolbar;
        this.popupMenu = popupMenu;
        this.showToast = showToast;
        this.dismissDialogWindow = dismissDialogWindow;

        // for being able to handle native (android) back button
        this.notifyBackPress = onBackPress;

        // utilities
        this.isHexColorDark = isHexColorDark;
        this.hexToColor = hexToColor;
        this.hexToInvertedRGB = hexToInvertedRGB;
        this.getADLVersion = getADLVersion;
        this.getIoniconVersion = getIoniconVersion;
        this.getScriptURL = getScriptURL;
        this.addADLStyle = addADLStyle;
        this.addIonicons = addIonicons;
        this.initADL = initADL;
        this.getOverallThemeLight = getOverallThemeLight;
        this.gamepadXYCheck = gamepadXYCheck;
        this.canElementBeGamepadSelected = canElementBeGamepadSelected;
        this.getDistance = getDistance;
        this.hexToOffset = hexToOffset;
        this.rgbToHex = rgbToHex;
    }

    function initADL() {

        if(adlIsInit) {
            return;
        }

        try {
            let lowerAgent = navigator.userAgent.toLowerCase();

            if(lowerAgent.indexOf("android") > -1) {
                deviceSpecial = "android";
            } else {
                if(lowerAgent.indexOf("iphone") > -1 || lowerAgent.indexOf("ipad") > -1) {
                    deviceSpecial = "ios";
                }

                if(lowerAgent.indexOf("macintosh") > -1) {
                    deviceSpecial = "macos";
                }
            }

        } catch(ex) {
            console.log(ex);
        }

        if(deviceSpecial) {
            if(deviceSpecial == "android") {
                document.body.classList.add("adlAndroid");
            }
            
            if(deviceSpecial == "ios") {
                document.body.classList.add("adliOS");
            }

            if(deviceSpecial == "macos") {
                document.body.classList.add("adlMacos");
            }
        }

        document.body.classList.add("adl");

        adlDynamicStyle = document.createElement("style");
        document.getElementsByTagName("head")[0].appendChild(adlDynamicStyle);

        gooseAllToolbars();

        setupDynamicStyles();

        adlIsInit = true;

    }

    function setupDynamicStyles() {
        if(!adlDynamicStyle) {
            return;
        }

        overallThemeLight = true;

        if(forceTheme) {
            if(forceTheme == "dark") {
                overallThemeLight = false;
            }
        } else {
            if (window.matchMedia && window.matchMedia("(prefers-color-scheme: dark)").matches) {
                overallThemeLight = false;
            }
        }

        if(overallThemeLight) {
            document.body.classList.remove("adlDark");
        } else {
            document.body.classList.add("adlDark");
        }

        adlDynamicStyle.innerHTML = "";

        let generalForeground = "#000000";
        let generalAccent = adlTheme;

        let darkAccent = hexToOffset(generalAccent,-40);
        let darkRGB = "rgb(" + darkAccent.r + "," + darkAccent.g + "," + darkAccent.b + ")";
        let lightForeground = isHexColorDark(adlTheme);
        
        let codeForeground = hexToOffset(generalAccent,-180);

        if(lightForeground) {
            generalForeground = "#ffffff";
            codeForeground = hexToOffset(generalAccent,180);
        }


        let codeForegroundRGB = "rgb(" + codeForeground.r + "," + codeForeground.g + "," + codeForeground.b + ")";

        // links
        adlDynamicStyle.innerHTML += "a{ color: " + generalAccent + "; } ";

        // buttons 
        adlDynamicStyle.innerHTML += "button.adl{ color: " + generalForeground + "; background-color: " + generalAccent + "; } ";

        // code box 
        adlDynamicStyle.innerHTML += ".adl-codeblock{ color: " + codeForegroundRGB + "; background-color: " + generalAccent + "; border-color: " + darkRGB + "; } ";

        // radio buttons
        adlDynamicStyle.innerHTML += ".adl-radioholder input:checked ~ .adl-radio { border-color: " + generalAccent + "; } .adl-radioholder .adl-radio:after { background: " + generalAccent + "; } ";

        // progress bars
        adlDynamicStyle.innerHTML += "progress.adl[value]::-webkit-progress-value { background-color: " + generalAccent + "; } ";

        // scroll bars 
        adlDynamicStyle.innerHTML += "*::-webkit-scrollbar-thumb { background-color: " + generalAccent + "; } ";

        // android text input border 
        adlDynamicStyle.innerHTML += ".adlAndroid input.adl[type=password],.adlAndroid input.adl[type=text],.adlAndroid textarea.adl { border-bottom: 2px solid " + generalAccent + "; } ";

        // toolbar 
        adlDynamicStyle.innerHTML += ".adl-toolbar-button .adl-toolbar-button-ionic { color: " + generalAccent + "; } .adl-toolbar-button .adl-toolbar-button-winicon { color: " + generalAccent + "; }";

        // android toolbar 
        adlDynamicStyle.innerHTML += ".adlAndroid .adl-toolbar { background-color: " + generalAccent + "; color: " + generalForeground + "; } .adlAndroid .adl-toolbar-button .adl-toolbar-button-ionic { color: " + generalForeground + "; } .adlAndroid .adl-toolbar-button .adl-toolbar-button-winicon { color: " + generalForeground + "; } ";

        // iOS toolbar 
        adlDynamicStyle.innerHTML += ".adliOS .adl-toolbar-button  .adl-toolbar-button-label-noicon { color: " + generalAccent + "; }  ";

        // toast action button 
        adlDynamicStyle.innerHTML += ".adl-toast-action { color: " + generalAccent + "; } ";
    }

    function setTheme(color,forceThemeLightOrDark) {

        if(color) {
            adlTheme = color;
        }

        if(forceThemeLightOrDark) {
            forceTheme = forceThemeLightOrDark;
        }

        setupDynamicStyles();
    }

    function isHexColorDark(hex) {
        let color = hexToColor(hex);

        let colorValue = color.r + color.g + color.b;

        if (colorValue > 382) {
            return false;
        } else {
            return true;
        }
    }

    function hexToColor(hex) {
        if (hex.length == 6 || hex.length == 7) {
            let result = /^#?([a-f\d]{2})([a-f\d]{2})([a-f\d]{2})$/i.exec(hex);
            return result ? {
                a: 255,
                r: parseInt(result[1], 16),
                g: parseInt(result[2], 16),
                b: parseInt(result[3], 16)
            } : null;
        }

        if (hex.length == 8 || hex.length == 9) {
            let result = /^#?([a-f\d]{2})([a-f\d]{2})([a-f\d]{2})([a-f\d]{2})$/i.exec(hex);
            return result ? {
                a: parseInt(result[1], 16),
                r: parseInt(result[2], 16),
                g: parseInt(result[3], 16),
                b: parseInt(result[4], 16)
            } : null;
        }

        return {
            a: 255,
            r: 0,
            g: 0,
            b: 0
        };
    }

    function hexToInvertedRGB(hex) {
        let color = hexToColor(hex);

        color.r = 255 - color.r;
        color.g = 255 - color.g;
        color.b = 255 - color.b;

        if(color.r < 0) {
            color.r = 0;
        }

        if(color.g < 0) {
            color.g = 0;
        }

        if(color.b < 0) {
            color.b = 0;
        }

        return color;
    }

    function hexToOffset(hex, offset) {
        let color = hexToColor(hex);

        color.r += offset;
        color.g += offset;
        color.b += offset;

        if (offset > 0) {
            if (color.r > 255) {
                color.r = 255;
            }

            if (color.g > 255) {
                color.g = 255;
            }

            if (color.b > 255) {
                color.b = 255;
            }
        } else {
            if (color.r < 0) {
                color.r = 0;
            }

            if (color.g < 0) {
                color.g = 0;
            }

            if (color.b < 0) {
                color.b = 0;
            }
        }

        return color;
    }

    function showDialog(options) {
        if(currentADLDialog) {

            setTimeout(function(){
                showDialog(options);
            },250);

            return;
        }

        pendingInputText = "";
        let inputElement = null;

        let gridTemplate = "";

        let flyoutFrom = null;
        let has1fr = false;

        document.body.classList.add("adlDialogOpen");

        currentADLDialog = options;

        prepareBlocker();

        let dialog = document.createElement("div");
        dialog.className = "adl adlDialog";

        if(options.backgroundColor) {
            dialog.style.backgroundColor = options.backgroundColor;

            let foreGround = "#000000";

            if(isHexColorDark(options.backgroundColor)) {
                foreGround = "#ffffff";
            }

            dialog.style.color = foreGround;
        }

        if(options.flyout) {
            if(options.flyout.position) {

                flyoutFrom = options.flyout.position;

                if(flyoutFrom == "bottom") {
                    dialog.className += " adlFlyoutBottom";
                }

                if(flyoutFrom == "top") {
                    dialog.className += " adlFlyoutTop";
                }

                if(flyoutFrom == "left") {
                    dialog.className += " adlFlyoutLeft";
                }

                if(flyoutFrom == "right") {
                    dialog.className += " adlFlyoutRight";
                }
            }
        }

        if(options.title) {
            let title = document.createElement("div");
            title.className = "adlDialogTitle";
            title.innerHTML = options.title;
            dialog.appendChild(title);

            gridTemplate += "auto ";
        }

        if(options.icon) {

            let imageURL = options.icon;

            if(isThisAnImage(imageURL)) {
                let icon = document.createElement("img");
                icon.className = "adlDialogImageIcon";
                icon.src = imageURL;
                icon.draggable = false;
                dialog.appendChild(icon);
            } else {
                let iconHolder = document.createElement("div");
                iconHolder.className = "adlDialogIonIcon";

                if(options.iconColor) {
                    iconHolder.style.color = options.iconColor;
                }

                let icon = document.createElement("ion-icon");
                icon.name = imageURL;
                iconHolder.appendChild(icon);

                dialog.appendChild(iconHolder);
            }
 
            if(deviceSpecial == "ios" || deviceSpecial == "macos") {
                gridTemplate += "auto ";
            }
        }

        if(options.message) {
            let message = document.createElement("div");
            message.className = "adlDialogMessage";
            message.innerHTML = replaceAll(options.message,"\n","<br />");

            if(!options.icon && deviceSpecial != "ios" && deviceSpecial != "macos") {
                message.style.gridColumn = "1";
            }

            dialog.appendChild(message);

            gridTemplate += "1fr ";
            has1fr = true;
        }

        if(options.input) {
            

            if(options.input.multiline) {
                inputElement = document.createElement("textarea");
            } else {
                inputElement = document.createElement("input");

                if(options.input.password) {
                    inputElement.type = "password";
                } else {
                    inputElement.type = "text";
                }
            }

            if(options.input.placeholder) {
                inputElement.placeholder = options.input.placeholder;
            }

            inputElement.oninput = function(){
                pendingInputText = this.value;
            };

            inputElement.className = "adl adlDialogInput adlGamepadSelectable";

            dialog.appendChild(inputElement);

            gridTemplate += "auto ";
        }

        if(options.slider) {
            let min = 0;
            let max = 100;
            let value = 1;
            let status = "1";

            if(options.slider.min) {
                min = options.slider.min;
            }

            if(options.slider.max) {
                max = options.slider.max;
            }

            if(options.slider.value) {
                value = options.slider.value;
            }

            if(options.slider.status) {
                status = options.slider.status;
            }

            pendingSliderValue = value;

            let sliderHolder = document.createElement("div");
            sliderHolder.className = "adlDialogSliderHolder";

            let slider = document.createElement("input");
            slider.type = "range";
            slider.className = "adl";

            slider.min = min;
            slider.max = max;
            slider.value = value;

            sliderHolder.appendChild(slider);

            let sliderStatus = document.createElement("div");
            sliderStatus.className = "adlDialogSliderStatus";
            sliderStatus.innerHTML = status;
            sliderHolder.appendChild(sliderStatus);

            if(options.slider.onChange) {
                slider.oninput = function(){
                    let val = parseInt(slider.value);

                    pendingSliderValue = val;

                    options.slider.onChange(val,function(text){
                        sliderStatus.innerHTML = text;
                    });
                };
            }

            dialog.appendChild(sliderHolder);

            gridTemplate += "auto ";
        }

        if(options.customElement) {
            let elementHolder = document.createElement("div");
            elementHolder.className = "adlDialogCustomElementHolder";
            elementHolder.appendChild(options.customElement);

            //elementHolder.

            dialog.appendChild(elementHolder);

            gridTemplate += "1fr ";
            has1fr = true;
        }



        if(options.progress) {
            let progressBar = document.createElement("progress");
            progressBar.className = "adl adlDialogProgress";
            
            if(options.progress.indeterminate) {
                progressBar.indeterminate = true;
            } else {
                progressBar.max = 100;

                if(options.progress.percent) {
                    progressBar.value = options.progress.percent;
                    progressBar.innerHTML = options.progress.percent + "%";
                } else {
                    progressBar.value = 0;
                    currentADLDialog.progress.percent = 0;
                }
            }

            dialog.appendChild(progressBar);

            gridTemplate += "auto ";
        }




        if(!options.buttons || options.buttons.length == 0) {
            options.buttons = [];

            if(!options.disableButtons && !options.flyout) {
                options.buttons.push({
                    text: "Dismiss"
                });
            }
            
        }

        if(!has1fr && flyoutFrom && (flyoutFrom == "left" || flyoutFrom == "right")) {
            let spacer = document.createElement("div");
            spacer.className = "adlFlyoutSpacerBlank";
            gridTemplate += "1fr ";
            dialog.appendChild(spacer);
        }

        if(options.buttons.length > 0) {

            let buttonColumns = "";

            let buttonsHolder = document.createElement("div");
            buttonsHolder.className = "adlDialogButtonsArea";

            if(options.buttons.length > 2) {
                buttonsHolder.className += " adlDialogButtonsAreaOvertwo";
            }
    
            for(let i = 0; i < options.buttons.length; i++) {
                let btnDef = options.buttons[i];
    
                let button = document.createElement("button");
                button.className = "adl adlGamepadSelectable";

                if(i > 0) {
                    button.className += " adlIosPadLeft";
                }

                button.innerHTML = btnDef.text;
    
                if(btnDef.color) {

                    if(deviceSpecial == "ios" || deviceSpecial == "android") {
                        button.style.color = btnDef.color;

                        if(deviceSpecial == "ios") {
                            button.style.fontWeight = "bold";
                        }
                        button.style.fontWeight = "bold";
                    } else {
                        let fg = "#000000";
                        let lightForeground = isHexColorDark(btnDef.color);
        
                        if(lightForeground) {
                            fg = "#ffffff";
                        }
        
                        button.style.backgroundColor = btnDef.color;
                        button.style.color = fg;
                    }

                    
                } else {
                    if(deviceSpecial == "ios" || deviceSpecial == "android") {
                        button.style.color = adlTheme;
                    } 
                }
    
                if(btnDef.func) {
                    button.funcRef = btnDef.func;
                }
    
                if(btnDef.inputResult) {
                    button.inputResult = btnDef.inputResult;
                }
    
                if(btnDef.sliderResult) {
                    button.sliderResult = btnDef.sliderResult;
                }
    
                button.oncontextmenu = killContext;
                button.onclick = dialogButtonClick;
    
                buttonsHolder.appendChild(button);

                buttonColumns += "1fr ";
            }

            if(deviceSpecial == "ios" || (deviceSpecial == "macos" && options.buttons.length < 3)) {
                buttonsHolder.style.gridTemplateColumns = buttonColumns.trim();
            }
    
            dialog.appendChild(buttonsHolder);

            gridTemplate += "auto ";
        }
        

        dialog.oncontextmenu = killContext;

        dialog.style.gridTemplateRows = gridTemplate.trim();
        
        adlBlocker.appendChild(dialog);

        if(flyoutFrom) {
            adlBlocker.onclick = function(e) {

                if(e.target.className != "adlBlocker") {
                    return;
                }

                if(!currentADLDialog.buttons || currentADLDialog.buttons.length < 2) {
                    dismissDialogWindow(null);
                }
                
                if(currentADLDialog.input) {
                    dismissDialogWindow(null);
                }
            };
        } else {
            adlBlocker.onclick = null;
        }

        setTimeout(function() {
            dialog.style.opacity = "1.0";

            if(flyoutFrom) {
                if(flyoutFrom == "bottom") {
                    dialog.style.bottom = "0%";
                }

                if(flyoutFrom == "top") {
                    dialog.style.top = "0%";
                }

                if(flyoutFrom == "left") {
                    dialog.style.left = "0px";
                }

                if(flyoutFrom == "right") {
                    dialog.style.right = "0px";
                }
            } else {
                dialog.style.transform = "translate(-50%,-50%) scale(1.0)";
            }

            dialog.focus();

            if(inputElement) {
                inputElement.focus();
            }
        },10);
    }

    function prepareBlocker() {
        if(!adlBlocker) {
            adlBlocker = document.createElement("div");
            adlBlocker.className = "adlBlocker";

            adlBlocker.oncontextmenu = function(e) {
                clearAndStopEvent(e);
            };



            document.body.appendChild(adlBlocker);
        }

        adlBlocker.innerHTML = "";
        adlBlocker.style.display = "block";
    }

    function isThisAnImage(url) {

        let imageURL = url.toLowerCase();

        if(
            imageURL.indexOf(".png") == -1 && 
            imageURL.indexOf(".jpg") == -1 && 
            imageURL.indexOf(".gif") == -1 &&
            imageURL.indexOf(".jpeg") == -1 &&
            imageURL.indexOf(".bmp") == -1 &&
            imageURL.indexOf(".tiff") == -1 &&
            imageURL.indexOf(".webp") == -1 &&
            imageURL.indexOf("data:image") == -1 && 
            imageURL.indexOf("gettoken.php") == -1 && 
            imageURL.indexOf("getavatar.php") == -1 && 
            imageURL.indexOf("geticon.php") == -1
        ) {
            return false;
        }

        return true;
    }

    function killContext(e) {
        e.preventDefault();
        e.stopPropagation();
    }

    function dialogButtonClick() {
        let funcRef = this.funcRef;
        let callback = funcRef;

        if(this.inputResult) {
            callback = function(){
                funcRef(pendingInputText);
            };
        }

        if(this.sliderResult) {
            callback = function(){
                funcRef(pendingSliderValue);
            };
        }

        dismissDialogWindow(callback);


    }

    function dismissDialogWindow(callback) {

        document.body.classList.remove("adlDialogOpen");

        if(!adlBlocker) {
            currentADLDialog = null;

            if(callback) {
                callback();
            }

            

            return false;
        }

        let dialog = adlBlocker.querySelector(".adlDialog");

        if(dialog) {
            dialog.style.opacity = "0.0";
        }

        setTimeout(function() {
            currentADLDialog = null;

            adlBlocker.innerHTML = "";
            adlBlocker.style.display = "none";


            if(callback) {
                callback();
            }

            return true;

        },150);
    }

    function onBackPress() {
        if(currentADLDialog) {
            if(!currentADLDialog.buttons || currentADLDialog.buttons.length < 2) {
                dismissDialogWindow(null);
                return true;
            }

            if(currentADLDialog.buttons && currentADLDialog.buttons.length >= 2) {
                // block back, dont dismiss
                return true;
            }
            
            if(currentADLDialog.input) {
                dismissDialogWindow(null);
                return true;
            }
        }

        // finally, check for popup menus
        return dismissPopupMenus();
    }

    function onKeyDown(e) {

        if(currentADLDialog) {
            if(e.key && e.key == "Escape") {
                if(!currentADLDialog.buttons || currentADLDialog.buttons.length < 2) {
                    clearAndStopEvent(e);
                    dismissDialogWindow(null);
                }
                
                if(currentADLDialog.input) {
                    clearAndStopEvent(e);
                    dismissDialogWindow(null);
                }
            }

            if(e.key && (e.key == "Enter" || e.key == "Return")) {
                let dialog = adlBlocker.querySelector(".adlDialog");
                let buttons = dialog.querySelectorAll("button");

                if(currentADLDialog.input) {

                    if(!currentADLDialog.input.multiline || e.ctrlKey) {
                        let funcRef = null;

                        
    
                        for(let i = 0; i < buttons.length; i++) {
                            let btn = buttons[i];
    
                            if(btn.inputResult) {
                                funcRef = btn.funcRef;
                            }
                        }
    
                        dismissDialogWindow(function() {
                            if(funcRef) {
                                funcRef(pendingInputText);
                            } else {
                                //dismissDialogWindow(null);
                            }
                            
                        });

                        clearAndStopEvent(e);
                    }
                    
                    return;
                }

                if(buttons.length > 0) {
                    clearAndStopEvent(e);

                    if(buttons.length == 1) {
                        let funcRef = null;

                        if(buttons[0].funcRef) {
                            funcRef = buttons[0].funcRef;
                        }

                        dismissDialogWindow(funcRef);
                    }


                    return;
                }
            }
                
        }
        

        
    }

    function showAlert(text) {
        showDialog({
            message: text
        });
    }

    function showAlertWithTitle(text,title) {
        showDialog({
            title: title,
            message: text
        });
    }

    function showChoice(text,b1,b2,f1,f2) {
        showDialog({
            message: text,
            buttons: [
                {
                    text: b1,
                    func: f1
                },
                {
                    text: b2,
                    func: f2
                }
            ]
        });
    }

    function inputBox(options) {

        showDialog({
            title: options.title,
            message: options.message,
            input: {
                multiline: options.multiline,
                password: options.password,
                placeholder: options.placeholder,
            },
            buttons: [
                {
                    text: "Cancel"
                },
                {
                    text: "Ok",
                    func: options.func,
                    inputResult: true
                }
            ]
        });
    }

    function replaceAll(str, find, replace) {
        return str.replace(new RegExp(find, "g"), replace);
    }

    function createColorSwatches(callback,colors,withInput,defColor) {

        if(!colors) {
            colors = DEFAULT_COLORPICKER_OPTIONS;
        }

        let colorsPerRow = 0;

        for(let i = 0; i < colors.length; i++) {
            let row = colors[i];

            if(row.length > colorsPerRow) {
                colorsPerRow = row.length;
            }
        }

        let pickerHolder = document.createElement("div");
        pickerHolder.className = "adlColorPicker";

        let colTemp = "";

        for(let n = 0; n < colorsPerRow; n++) {
            colTemp += "1fr ";
        }

        pickerHolder.style.gridTemplateColumns = colTemp.trim();

        let allSwatches = [];

        for(let i = 0; i < colors.length; i++) {
            let row = colors[i];

            for(let j = 0; j < row.length; j++) {
                let color = row[j];

                let swatch = document.createElement("div");
                swatch.className = "adlColorPickerSwatch adlGamepadSelectable";
                swatch.colorRef = color;
                swatch.callbackRef = callback;
                swatch.onclick = onSwatchSelected;
                swatch.style.backgroundColor = color;

                pickerHolder.appendChild(swatch);
                allSwatches.push(swatch);
            }

        }

        if(withInput) {
            let inputRow = document.createElement("div");
            inputRow.style.gridColumn = "1 / span " + colorsPerRow;
            inputRow.className = "adlColorPickerInputHolder";

            let inputHexHeader = document.createElement("div");
            inputHexHeader.className = "adlColorPickerHexHeader";
            inputHexHeader.innerHTML = "Hex";
            inputRow.appendChild(inputHexHeader);

            let inputHexTextInput = document.createElement("input");
            inputHexTextInput.type = "text";
            inputHexTextInput.className = "adl";
            inputHexTextInput.placeholder = "Hex";

            

            if(defColor) {
                inputHexTextInput.value = defColor;
            }

            inputRow.appendChild(inputHexTextInput);

            let inputInput = document.createElement("input");
            inputInput.type = "color";

            if(defColor) {
                inputInput.value = defColor;
            }

            inputInput.oninput = function(){
                callback(this.value);
                inputHexTextInput.value = this.value;
            };

            inputHexTextInput.oninput = function(){
                let val = this.value.trim();

                if(val.length == 7 && val.indexOf("#") == 0) {
                    callback(val);
                    inputInput.value = val;
                }
            };

            inputRow.appendChild(inputInput);

            pickerHolder.appendChild(inputRow);

            for(let i = 0; i < allSwatches.length; i++) {
                let sr = allSwatches[i];
                sr.textInputRef = inputHexTextInput;
                sr.colorInputRef = inputInput;
            }
        }

        return pickerHolder;
    }

    function onSwatchSelected() {
        let color = this.colorRef;
        let callback = this.callbackRef;

        callback(color);

        if(this.textInputRef) {
            this.textInputRef.value = color;
        }

        if(this.colorInputRef) {
            this.colorInputRef.value = color;
        }
    }

    function pickColor(options) {

        let colors = options.colors;

        if(options.default) {
            pendingColorChoice = options.default;
        } else {
            pendingColorChoice = null;
        }

        let pickerHolder = createColorSwatches(function(color) {

            let shouldDismiss = false;

            if(!options.element) {
                if(!options.withInput) {
                    shouldDismiss = true;
                }
            }

            if(shouldDismiss) {
                dismissDialogWindow(null);
                options.callback(color);
            } else {
                pendingColorChoice = color;

                if(options.element) {
                    options.callback(color);
                }
            }
            
            
        },colors,options.withInput,options.default);

        if(options.element) {
            options.element.innerHTML = "";
            options.element.appendChild(pickerHolder);
        } else {

            let pickerButtons = [{
                text: "Cancel"
            }];

            if(options.withInput) {
                pickerButtons.push({
                    text: "Pick",
                    func: function(){
                        options.callback(pendingColorChoice);
                    }
                });
            }

            showDialog({
                title: "Pick Color",
                customElement: pickerHolder,
                buttons: pickerButtons
            });
        }

        
    }

    function clearAndStopEvent(e) {
        e.preventDefault();
        e.stopPropagation();
    }

    function showProgress(options) {

        if(!options.title) {
            options.title = "Please Wait...";
        }

        if(!options.indeterminate) {
            options.indeterminate = false;
        }

        if(!options.percent) {
            options.percent = 0;
        }

        showDialog({
            title: options.title,
            progress: {
                indeterminate: options.indeterminate,
                percent: options.percent
            },
            disableButtons: true
        });
    }

    function updateProgress(options) {

        if(!currentADLDialog || !currentADLDialog.progress) {
            return;
        }

        let dialog = adlBlocker.querySelector(".adlDialog");

        if(!dialog) {
            return;
        }

        let progressBar = dialog.querySelector(".adlDialogProgress");

        if(!progressBar) {
            return;
        }

        let title = dialog.querySelector(".adlDialogTitle");

        if(options.indeterminate == undefined) {
            options.indeterminate = currentADLDialog.progress.indeterminate;
        }

        if(options.percent == undefined) {
            options.percent = currentADLDialog.progress.percent;
        }

        if(!options.title) {
            options.title = currentADLDialog.title;
        }

        currentADLDialog.title = options.title;
        currentADLDialog.progress.percent = options.percent;
        currentADLDialog.progress.indeterminate = options.indeterminate;

        if(title) {
            title.innerHTML = options.title;
        }

        progressBar.indeterminate = options.indeterminate;

        if(!progressBar.indeterminate) {
            progressBar.value = options.percent;
        }
    }

    function dismissProgress() {
        dismissDialogWindow(null);
    }

    function showList(options) {
        if(!options.title) {
            options.title = "Choose Something";
        }

        if(!options.cancelTest && !options.flyout) {
            options.cancelTest = "Dismiss";
        }

        let flyout;

        if(options.flyout) {
            flyout = options.flyout;
        }

        if(options.element) {
            let list = createScrollingList(options.options,options.onSelection);

            options.element.innerHTML = "";
            options.element.appendChild(list);
        } else {
            let list = createScrollingList(options.options,function(tag) {
                dismissDialogWindow(null);
                options.onSelection(tag);
            });

            let listButtons = [];

            if(options.cancelTest) {
                listButtons.push({
                    text: options.cancelTest,
                    func: function() {
                        options.onSelection(null);
                    }
                });
            }
            

            showDialog({
                title: options.title,
                customElement: list,
                buttons: listButtons,
                flyout: flyout
            });
        }
    }

    function createScrollingList(options,callback) {
        let list = document.createElement("div");
        list.className = "adl-list";
        list.ontouchstart = function(e) {
            e.stopPropagation();
        };

        for(let i = 0; i < options.length; i++) {
            let option = options[i];

            let item = document.createElement("div");
            item.className = "adl-list-item adlGamepadSelectable";

            if(option.compactItem) {
                item.className += " adl-compact-list-item";
            }

            if(option.icon) {
                item.className += " adl-list-item-withIcon";
                let imageURL = option.icon;

                if(isThisAnImage(imageURL)) {
                    let icon = document.createElement("img");
                    icon.className = "adl-list-item-imageIcon";
                    icon.src = imageURL;
                    icon.draggable = false;
                    item.appendChild(icon);
                } else {
                    let iconHolder = document.createElement("div");

                    if(option.iconColor) {
                        iconHolder.style.color = option.iconColor;
                    }

                    if(imageURL.indexOf("winicon.") == 0) {
                        iconHolder.className = "adl-list-item-winIcon";
    
                        let icon = document.createElement("span");
                        icon.className = "winIconInner";
                        icon.innerHTML = imageURL.replace("winicon.","");
                        iconHolder.appendChild(icon);
                    } else {
                        iconHolder.className = "adl-list-item-ionIcon";

                        let icon = document.createElement("ion-icon");
                        icon.name = imageURL;
                        iconHolder.appendChild(icon);
                    }

                    

                    item.appendChild(iconHolder);
                }
            }

            let contentHolder = document.createElement("div");
            contentHolder.className = "adl-list-item-content";

            if(option.title) {
                let itemTitle = document.createElement("div");
                itemTitle.className = "adl-list-item-title";
                itemTitle.innerHTML = option.title;

                if(option.titleColor) {
                    itemTitle.style.color = option.titleColor;
                }

                contentHolder.appendChild(itemTitle);
            }

            if(option.description) {
                let itemDesc = document.createElement("div");
                itemDesc.className = "adl-list-item-description";
                itemDesc.innerHTML = option.description;
                contentHolder.appendChild(itemDesc);
            }

            if(option.progress != undefined) {

                if(isNaN(option.progress)) {
                    option.progress = 0;
                }

                let progressBar = document.createElement("progress");
                progressBar.className = "adl adl-list-item-progress";
                progressBar.max = 100;
                progressBar.value = option.progress;
                contentHolder.appendChild(progressBar);
            }

            if(option.status) {
                let itemStatus = document.createElement("div");
                itemStatus.className = "adl-list-item-status";
                itemStatus.innerHTML = option.status;

                if(option.statusColor) {
                    itemStatus.style.color = option.statusColor;
                }

                contentHolder.appendChild(itemStatus);
            }

            if(option.overflow && option.overflow.length > 0) {
                for(let j = 0; j < option.overflow.length; j++) {
                    let ofItem = option.overflow[j];

                    let overflowHolder = document.createElement("div");
                    overflowHolder.className = "adl-list-overflow-item";

                    if(ofItem.color) {
                        overflowHolder.style.color = ofItem.color;
                    }

                    if(ofItem.pre) {
                        let preItem = document.createElement("div");
                        preItem.className = "adl-list-overflow-item-pre";
                        preItem.innerHTML = ofItem.pre;
                        overflowHolder.appendChild(preItem);
                    }

                    if(ofItem.icon) {
                        let ofImageURL = ofItem.icon;

                        if(isThisAnImage(ofImageURL)) {
                            let icon = document.createElement("img");
                            icon.className = "adl-list-overflow-item-imageIcon";
                            icon.src = ofImageURL;
                            icon.draggable = false;
                            overflowHolder.appendChild(icon);
                        } else {
                            let iconHolder = document.createElement("div");
                            iconHolder.className = "adl-list-overflow-item-ionIcon";

                            if(ofItem.iconColor) {
                                iconHolder.style.color = ofItem.iconColor;
                            }

                            let icon = document.createElement("ion-icon");
                            icon.name = ofImageURL;
                            iconHolder.appendChild(icon);

                            overflowHolder.appendChild(iconHolder);
                        }
                    }

                    if(ofItem.post) {
                        let postItem = document.createElement("div");
                        postItem.className = "adl-list-overflow-item-post";
                        postItem.innerHTML = ofItem.post;
                        overflowHolder.appendChild(postItem);
                    }

                    contentHolder.appendChild(overflowHolder);
                }
            }

            item.appendChild(contentHolder);

            item.optionRef = option;
            item.callbackRef = callback;
            item.onclick = onListItemClick;
            item.oncontextmenu = clearAndStopEvent;

            list.appendChild(item);
        }

        return list;
    }

    function onListItemClick() {

        let option = this.optionRef;
        let callback = this.callbackRef;

        if(!option || !callback) {
            return;
        }

        if(option.tag) {
            callback(option.tag);
        } else {
            callback(null);
        }
    }

    function gooseAllToolbars() {
        let toolbars = document.querySelectorAll(".adl-toolbar");

        for(let i = 0; i < toolbars.length; i++) {
            let toolbar = toolbars[i];

            setupToolbar(toolbar);
        }
    }

    function setupToolbar(toolbar) {
        if(!toolbar) {
            return;
        }

        let items = toolbar.querySelectorAll(".adl-toolbar-item");

        let gridDef = "";

        let forceLabel = false;

        if(toolbar.dataset && toolbar.dataset.forcelabel) {
            forceLabel = true;
        }

        for(let i = 0; i < items.length; i++) {
            let item = items[i];

            item.oncontextmenu = clearAndStopEvent;

            let data = item.dataset;

            let type = data.type;

            if(!type) {
                item.style.display = "none";
                continue;
            }

            if(type == "title") {
                gridDef += "1fr ";
                item.className += " adl-toolbar-title";

                if(data.label) {
                    item.innerHTML = data.label;
                }
                
            }

            if(type == "spacer") {
                gridDef += "1fr ";
                item.className += " adl-toolbar-spacer";
            }

            if(type == "button") {
                gridDef += "auto ";
                item.className += " adl-toolbar-button";

                let useIcon = null;

                if(data.icon) {
                    useIcon = data.icon;
                }

                if(data.special) {
                    if(data.special == "hamburger") {
                        useIcon = "menu-outline";

                        if(deviceSpecial && deviceSpecial == "android") {
                            useIcon = "menu-sharp";
                        }
    
                        if(deviceSpecial && deviceSpecial == "ios") {
                            useIcon = "ellipsis-horizontal-outline";
                        }
                    }

                    if(data.special == "threedots") {
                        useIcon = "ellipsis-vertical-outline";

                        if(deviceSpecial && deviceSpecial == "android") {
                            useIcon = "ellipsis-vertical";
                        }
    
                        if(deviceSpecial && deviceSpecial == "ios") {
                            useIcon = "ellipsis-vertical-circle-outline";
                        }
                    }
                }

                if(useIcon) {
                    if(isThisAnImage(useIcon)) {
                        let icon = document.createElement("img");
                        icon.className = "adl-toolbar-button-img";

                        if(data.circleIcon) {
                            icon.className += " adl-toolbar-button-img-circle";
                        }

                        icon.src = useIcon;
                        icon.draggable = false;
                        item.appendChild(icon);
                    } else {
    
                        
    
                        let iconHolder = document.createElement("div");

                        if(data.color) {
                            iconHolder.style.color = data.color;
                        }

                        if(data.iconColor) {
                            iconHolder.style.color = data.iconColor;
                        }

                        if(useIcon.indexOf("winicon.") == 0) {
                            iconHolder.className = "adl-toolbar-button-winicon";

                            let icon = document.createElement("span");
                            icon.className = "winIconInner";
                            icon.innerHTML = useIcon.replace("winicon.","");

                            if(data.label) {
                                icon.title = data.label;
                            }
                            
                            iconHolder.appendChild(icon);
                        } else {
                            iconHolder.className = "adl-toolbar-button-ionic";

                       

                            let icon = document.createElement("ion-icon");
                            icon.name = useIcon;

                            if(data.label) {
                                icon.title = data.label;
                            }

                            iconHolder.appendChild(icon);
                        }

                        if(data.label) {
                            iconHolder.title = data.label;
                        }
        
                        item.appendChild(iconHolder);
                    }
                }
                

                if(data.label) {
                    item.title = data.label;
                    
                    let labelSpan = document.createElement("div");
                    labelSpan.className = "adl-toolbar-button-label";
                    labelSpan.innerHTML = data.label;

                    if(data.color) {
                        labelSpan.style.color = data.color;
                    }

                    if(!useIcon) {
                        labelSpan.className += " adl-toolbar-button-label-noicon";
                    }

                    if(deviceSpecial && forceLabel && (deviceSpecial == "android" || deviceSpecial == "ios")) {
                        labelSpan.style.display = "block";
                    }

                    item.appendChild(labelSpan);
                }
            }
        }

        toolbar.style.gridTemplateColumns = gridDef.trim();
    }

    function addToolbar(options) {

        if(!options.element) {
            return;
        }

        options.element.innerHTML = "";

        let toolbar = document.createElement("div");
        toolbar.className = "adl-toolbar";

        if(options.background) {
            toolbar.style.backgroundColor = options.background;
        }

        if(options.foreground) {
            toolbar.style.color = options.foreground;
        }

        if(options.forcelabel) {
            toolbar.dataset.forcelabel = true;
        }

        if(options.items) {
            for(let i = 0; i < options.items.length; i++) {
                let itemDef = options.items[i];

                let item = document.createElement("div");
                item.className = "adl-toolbar-item";

                if(itemDef.type) {
                    item.dataset.type = itemDef.type;
                }

                if(itemDef.icon) {
                    item.dataset.icon = itemDef.icon;
                }

                if(itemDef.circleIcon) {
                    item.dataset.circleIcon = itemDef.circleIcon;
                }

                if(itemDef.special) {
                    item.dataset.special = itemDef.special;
                }

                if(itemDef.label) {
                    item.dataset.label = itemDef.label;
                }

                if(itemDef.iconColor) {
                    item.dataset.iconColor = itemDef.iconColor;
                }

                if(itemDef.tag) {
                    item.tagRef = itemDef.tag;
                }

                if(options.foreground) {
                    item.dataset.color = options.foreground;
                }

                if(itemDef.func || itemDef.menu) {
                    if(itemDef.func) {
                        item.cbRef = itemDef.func;
                    }

                    if(itemDef.menu) {
                        item.menuRef = itemDef.menu;
                        item.onmouseenter = onToolbarItemMouseOver;
                    }
                    
                    item.toolbarRef = toolbar;
                    item.onclick = onToolbarItemClick;
                }

                toolbar.appendChild(item);
            }
        }

        setupToolbar(toolbar);

        options.element.appendChild(toolbar);
    }

    function onToolbarItemClick(e) {

        dismissPopupMenus();

        let elemRect = this.getBoundingClientRect();

        let retData = {
            x: e.clientX,
            y: e.clientY,
            toolbarRect: elemRect
        };

        if(this.tagRef) {
            retData.tag = this.tagRef;
        }

        if(this.cbRef) {
            this.cbRef(retData);
        }

        if(this.menuRef) {

            let x = e.clientX;
            let y = e.clientY;

            x = elemRect.left;
            y = elemRect.bottom;

            popupMenu(this.menuRef,x,y,this.toolbarRef);
        }
    }

    function popupMenu(menu,x,y,fromToolbar) {

        if(!menu || !menu.items || menu.items.length == 0) {
            return;
        }

        if(fromToolbar) {
            fromToolbar.style.position = "relative";
            fromToolbar.style.zIndex = "998";
        }

        let menuElement = document.createElement("div");
        menuElement.toolbarRef = fromToolbar;
        menuElement.className = "adl-popup-menu";
        menuElement.menuRef = menu;

        

        if(!adlBlocker || adlBlocker.style.display == "none") {
            prepareBlocker();
        }

        document.body.classList.add("adlDialogOpen");

        let hasImages = false;

        for(let i = 0; i < menu.items.length; i++) {
            let item = menu.items[i];

            if(!item) {
                continue;
            }

            let itemElement = document.createElement("div");
            itemElement.className = "adl-popup-menu-item";
            itemElement.itemRef = item;
            itemElement.onclick = onPopupItemClick;

            if(item.tag) {
                itemElement.tagRef = item.tag;
            }
            
            if(item.func) {
                itemElement.funcRef = item.func;
            }

            if(item.type && (item.type == "separator" || item.type == "seperator")) {
                itemElement.className += " adl-popup-menu-item-separator";
            }

            if(item.icon) {
                let imageURL = item.icon;

                hasImages = true;

                if(isThisAnImage(imageURL)) {
                    let icon = document.createElement("img");
                    icon.className = "adl-popup-menu-item-imageIcon";
                    icon.src = imageURL;
                    icon.draggable = false;

                    itemElement.appendChild(icon);
                } else {
                    let iconHolder = document.createElement("div");

                    if(item.iconColor) {
                        iconHolder.style.color = item.iconColor;
                    }

                    if(imageURL.indexOf("winicon.") == 0) {
                        iconHolder.className = "adl-popup-menu-item-winIcon";

                        let icon = document.createElement("span");
                        icon.className = "winIconInner";
                        icon.innerHTML = imageURL.replace("winicon.","");
                        iconHolder.appendChild(icon);
                    } else {
                        iconHolder.className = "adl-popup-menu-item-ionIcon";

                        let icon = document.createElement("ion-icon");
                        icon.name = imageURL;
                        iconHolder.appendChild(icon);
                    }

                    itemElement.appendChild(iconHolder);
                }
            }

            if(item.title) {
                let titleDiv = document.createElement("div");
                titleDiv.className = "adl-popup-menu-item-title";
                titleDiv.innerHTML = item.title;

                if(item.color) {
                    titleDiv.style.color = item.color;
                }

                itemElement.appendChild(titleDiv);
            }

            if(item.accelerator) {
                let accelDiv = document.createElement("div");
                accelDiv.className = "adl-popup-menu-item-accelerator";
                accelDiv.innerHTML = item.accelerator;

                itemElement.appendChild(accelDiv);
            }

            menuElement.appendChild(itemElement);
        }

        if(!hasImages) {
            menuElement.className += " adl-popup-menu-no-images";
        }

        adlBlocker.appendChild(menuElement);

        adlBlocker.onclick = function(e) {

            if(e.target.className != "adlBlocker") {
                return;
            }

            dismissPopupMenus();
        };

        if(fromToolbar) {
            let elemRect = fromToolbar.getBoundingClientRect();

            if(elemRect) {
                if(y > window.innerHeight / 2) {
                    y = elemRect.top;
                }
            }
        }

        arrangeMenu(menuElement,x,y);

        setTimeout(function(){
            menuElement.style.opacity = "1";
            menuElement.style.transform = "initial";
        },10);

        setTimeout(function(){
            arrangeMenu(menuElement,x,y);
        },100);


    }

    function arrangeMenu(menuElement,x,y) {
        let mw = menuElement.offsetWidth;

        let w = window.innerWidth;
        let h = window.innerHeight;

        let xProp = "left";
        let yProp = "top";

        
        if(x > (w - mw)) {
            xProp = "right";
            x = 0;
        }

        if(y > h / 2) {
            yProp = "bottom";
            y = h - y;
        }

        menuElement.style[xProp] = x + "px";
        menuElement.style[yProp] = y + "px";
    }

    function dismissPopupMenus() {

        if(!adlBlocker) {
            return;
        }

        let menus = document.querySelectorAll(".adl-popup-menu");

        let gotOne = false;

        for(let i = 0; i < menus.length; i++) {
            let menu = menus[i];

            if(menu.toolbarRef) {
                menu.toolbarRef.style.position = null;
                menu.toolbarRef.style.zIndex = null;
            }

            nukePopupMenu(menu);

            gotOne = true;
            
        }

        document.body.classList.remove("adlDialogOpen");
        adlBlocker.innerHTML = "";
        adlBlocker.style.display = "none";

        return gotOne;
    }

    function nukePopupMenu(menu) {

        if(!menu) {
            return;
        }

        menu.style.opacity = 0;
        menu.style.transform = "scale(0)";

        setTimeout(function() {
            if(menu && menu.parentElement) {
                menu.innerHTML = "";
                menu.parentElement.removeChild(menu);
            }
            
        },100);
    }

    function onPopupItemClick() {
        dismissPopupMenus();

        if(this.funcRef) {
            this.funcRef(this.tagRef);
        }
    }

    function onToolbarItemMouseOver(e) {
        let thisButton = this;

        if(!this.menuRef) {
            return;
        }

        let menus = document.querySelectorAll(".adl-popup-menu");

        if(menus.length == 0) {
            return;
        }

        if(menus.length == 1) {
            if(menus[0].menuRef && menus[0].menuRef == thisButton.menuRef) {
                return; 
            }
        }

        dismissPopupMenus();

        let elemRect = this.getBoundingClientRect();

        let x = e.clientX;
        let y = e.clientY;

        if(elemRect) {
            x = elemRect.left;
            y = elemRect.bottom;
        }
        

        popupMenu(this.menuRef,x,y,this.toolbarRef);
    }

    function showToast(options) {

        if(!options) {
            return;
        }

        let currentToast = document.querySelector(".adl-toast");

        if(currentToast) {
            toastQueue.push(options);
            return;
        }

        let time = 3500;

        if(options.seconds) {
            time = options.seconds * 1000;
        }

        let toast = document.createElement("div");
        toast.className = "adl-toast";
        toast.onclick = dismissToast;

        if(options.background) {
            toast.style.backgroundColor = options.background;

            if(!isHexColorDark(options.background)) {
                toast.style.color = "#000000";
            }
        }

        if(options.icon) {
            let imageURL = options.icon;

            if(isThisAnImage(imageURL)) {
                let icon = document.createElement("img");
                icon.className = "adl-toast-imageIcon";
                icon.src = imageURL;
                icon.draggable = false;

                toast.appendChild(icon);
            } else {
                let iconHolder = document.createElement("div");

                if(options.iconColor) {
                    iconHolder.style.color = options.iconColor;
                }
                
                if(imageURL.indexOf("winicon.") == 0) {
                    iconHolder.className = "adl-toast-winIcon";

                    let icon = document.createElement("span");
                    icon.className = "winIconInner";
                    icon.innerHTML = imageURL.replace("winicon.","");
                    iconHolder.appendChild(icon);
                } else {
                    iconHolder.className = "adl-toast-ionIcon";

                    let icon = document.createElement("ion-icon");
                    icon.name = imageURL;
                    iconHolder.appendChild(icon);
                }

                toast.appendChild(iconHolder);
            }
        }

        if(options.message) {
            let toastMessage = document.createElement("div");
            toastMessage.innerHTML = options.message;
            toastMessage.className = "adl-toast-message";
            toast.appendChild(toastMessage);
        }

        if(options.action) {
            let toastAction = document.createElement("div");
            toastAction.innerHTML = options.action.text;
            toastAction.className = "adl-toast-action";
            toastAction.onclick = function(){
                if(options.action.func) {
                    options.action.func();
                }
            };

            if(options.action.color) {
                toastAction.style.color = options.action.color;
            }

            toast.appendChild(toastAction);
        }


        document.body.appendChild(toast);

        setTimeout(function() {
            toast.style.bottom = "12px";
            toast.style.transform = "translateX(-50%)";
        },50);

        toastDismissTimeout = setTimeout(dismissToast,time);
    }

    function dismissToast() {

        try {
            if(toastDismissTimeout) {
                clearTimeout(toastDismissTimeout);
            }
        } catch(ex) {
            console.log(ex);
        }

        toastDismissTimeout = null;

        let currentToast = document.querySelector(".adl-toast");

        if(currentToast) {
            currentToast.style.bottom = "-80px";
            currentToast.style.opacity = "0";

            setTimeout(function(){
                if(currentToast && currentToast.parentElement) {
                    currentToast.parentElement.removeChild(currentToast);
                    currentToast = null;
                    
                }

                nextToastIfAny();
            },250);
        } else {
            nextToastIfAny();
        }
    }

    function nextToastIfAny() {
        if(toastQueue && toastQueue.length > 0) {
            let next = toastQueue.shift();
            showToast(next);
        }
    }

    function getADLVersion() {
        return ADL_VERSION;
    }

    function getScriptURL(scriptName) {
        if(!scriptName) {
            scriptName = "adl.js";
        }

        let scripts = document.getElementsByTagName("script");

        let myScript = null;

        for(let i = 0; i < scripts.length; i++) {
            let script = scripts[i];
            if(script.src.indexOf("adl.js") > -1) {
                myScript = script;
                break;
            }
        }

        if(myScript) {
            return myScript.src;
        }

        return null;
    }

    function addADLStyle() {
        let adlScript = getScriptURL();

        if(adlScript) {
            let cssScript = adlScript.replace("js/adl.js","css/adl.css");

            let linkTag = document.createElement("link");
            linkTag.rel = "stylesheet";
            linkTag.type = "text/css";
            linkTag.href = cssScript;

            document.head.insertBefore(linkTag, document.head.firstChild);
        }
    }

    function addIonicons() {
        let s2Tag = document.createElement("script");
        s2Tag.noModule = "";
        s2Tag.src = "https://unpkg.com/ionicons@" + IONIC_VERSION + "/dist/ionicons/ionicons.js";
        document.head.insertBefore(s2Tag, document.head.firstChild);

        let s1Tag = document.createElement("script");
        s1Tag.type = "module";
        s1Tag.src = "https://unpkg.com/ionicons@" + IONIC_VERSION + "/dist/ionicons/ionicons.esm.js";
        document.head.insertBefore(s1Tag, document.head.firstChild);
    }

    function getIoniconVersion() {
        return IONIC_VERSION;
    }

    function getOverallThemeLight() {
        return overallThemeLight;
    }

    function getGamepadSelectableElements() {

        let checkEle = document;

        if(currentADLDialog) {
            checkEle = adlBlocker;
        }

        let allEles = checkEle.querySelectorAll(".adlGamepadSelectable");
        let selectableEles = [];

        for(let i = 0; i < allEles.length; i++) {
            let element = allEles[i];

            let visible = true;
            let parent = element;

            while(parent && visible) {
                if(parent.style.display == "none") {
                    visible = false;
                }

                parent = parent.parentElement;
            }

            if(visible) {
                selectableEles.push(element);
            }
        }

        return selectableEles;
    }

    function canElementBeGamepadSelected(element) {
        if(element) {
            let elements = getGamepadSelectableElements();

            if(elements.length > 0) {
                if(elements.indexOf(element) > -1) {
                    return true;
                }
            }
        }
        

        return false;
    }

    function gamepadXYCheck(direction,compareElement) {

        let nextElement = null;

        let elements = getGamepadSelectableElements();

        if(elements.length == 0) {
            return nextElement;
        }

        if(!compareElement || elements.indexOf(compareElement) == -1) {
            compareElement = elements[0];
            return compareElement;
        }

        nextElement = elements[0];

        let checkBounds = compareElement.getBoundingClientRect();
        let closestElement = 999999;

        for(let i = 0; i < elements.length; i++) {
            let element = elements[i];

            if(element == compareElement) {
                continue;
            }

            let bounds = element.getBoundingClientRect();

            let doCheck = false;

            if(direction == "right") {
                if(bounds.left > checkBounds.left) {
                    doCheck = true;
                }
            }

            if(direction == "left") {
                if(bounds.left < checkBounds.left) {
                    doCheck = true;
                }
            }

            if(direction == "up") {
                if(bounds.top < checkBounds.top) {
                    doCheck = true;
                }
            }

            if(direction == "down") {
                if(bounds.top > checkBounds.top) {
                    doCheck = true;
                }
            }

            if(doCheck) {
                let dist = getDistance(checkBounds.left,checkBounds.top,bounds.left,bounds.top);

                if(dist < closestElement) {
                    nextElement = element;
                    closestElement = dist;
                }
            }
        }

        return nextElement;
    }

    function getDistance(x1,y1,x2,y2) {
        let a = x1 - x2;
        let b = y1 - y2;
        let distance = Math.sqrt( a*a + b*b );

        return distance;
    }

    function rgbToHex(r, g, b) {
        return "#" + componentToHex(r) + componentToHex(g) + componentToHex(b);
    }

    function componentToHex(c) {
        let hex = c.toString(16);
        return hex.length == 1 ? "0" + hex : hex;
    }

})();