// reference: http://codepen.io/falldowngoboone/pen/zxRXjL
(function() {
    window.floodFill = floodFill;
    window.hexToRgb = hexToRgb;
    
    let tolerence = 30;
    let imageData;
    let startColor;
    
    function hexToRgb(hex) {
        let result = /^#?([a-f\d]{2})([a-f\d]{2})([a-f\d]{2})$/i.exec(hex);
        return result ? {
            r: parseInt(result[1], 16),
            g: parseInt(result[2], 16),
            b: parseInt(result[3], 16),
            a: 255
        } : null;
    }
    
    function floodFill(cx,startX,startY,color,tol) {

        if(tol) {
            tolerence = tol;
        }
        
        imageData = cx.getImageData(0,0,cx.canvas.width,cx.canvas.height);
        let startPixel = cx.getImageData(startX,startY,1,1);
        
        startColor = {};
        startColor.r = startPixel.data[0];
        startColor.g = startPixel.data[1];
        startColor.b = startPixel.data[2];
        startColor.a = startPixel.data[3];
        
        let fillColor = hexToRgb(color);
        
        let fillSameAsStart = true;
        
        if(fillColor.a < startColor.a - tolerence || fillColor.a > startColor.a + tolerence) {
            fillSameAsStart = false;
        }

        if(fillColor.r < startColor.r - tolerence || fillColor.r > startColor.r + tolerence) {
            fillSameAsStart = false;
        }
        
        if(fillColor.g < startColor.g - tolerence || fillColor.g > startColor.g + tolerence) {
            fillSameAsStart = false;
        }
        
        if(fillColor.b < startColor.b - tolerence || fillColor.b > startColor.b + tolerence) {
            fillSameAsStart = false;
        }
        
        if(fillSameAsStart) {
            //console.log("fill same as start!");
            return;
        }
        
        let canvasWidth = cx.canvas.width;
        let canvasHeight = cx.canvas.height;
        let drawingBoundTop = 0;
        let drawingBoundLeft = 0;
        let drawingBoundBottom = canvasHeight - 1;
        let drawingBoundRight = canvasWidth - 1;
        
        let pixelStack = [[startX, startY]];
        
        let newPos, x, y, pixelPos, reachLeft, reachRight, id, isPainted;

        isPainted = new Array(imageData.width * imageData.height);
        
        for(let i = 0; i < isPainted.length; i++) {
            isPainted[i] = false;
        }
        
        while(pixelStack.length) {
            //let newPos, x, y, pixelPos, reachLeft, reachRight;
            newPos = pixelStack.pop();
            x = newPos[0];
            y = newPos[1];

            pixelPos = (y * canvasWidth + x) * 4;
            
            while(y >= drawingBoundTop && matchStartColor(pixelPos)) {
                y -=1;
                pixelPos -= canvasWidth * 4;
            }
            
            pixelPos += canvasWidth * 4;
            y += 1;
            
            reachLeft = false;
            reachRight = false;

            //id = x + y * imageData.width;
            
            while(y < drawingBoundBottom && matchStartColor(pixelPos)) {
                
                y += 1;

                id = x + y * imageData.width;

                if (isPainted[id]) {
                    continue;
                } else {
                    imageData.data[pixelPos] = fillColor.r;
                    imageData.data[pixelPos+1] = fillColor.g;
                    imageData.data[pixelPos+2] = fillColor.b;
                    imageData.data[pixelPos+3] = fillColor.a;

                    isPainted[id] = true;
                }

                

                if(x > drawingBoundLeft) {
                    if(matchStartColor(pixelPos - 4)) {
                        if(!reachLeft) {
                            pixelStack.push([x - 1, y]);
                            reachLeft = true;
                        }
                    } else if (reachLeft) {
                        reachLeft = false;
                    }
                }

                if(x < drawingBoundRight) {
                    if(matchStartColor(pixelPos + 4)) {
                        if(!reachRight) {
                            pixelStack.push([x + 1, y]);
                            reachRight = true;
                        }
                    } else if(reachRight) {
                        reachRight = false;
                    }
                }
                    
                pixelPos += canvasWidth * 4;
            }
        }
        
        cx.putImageData(imageData, 0, 0);

        imageData = null;
        startColor = null;
    }
    
    function matchStartColor(pos) {
        let r = imageData.data[pos];    
        let g = imageData.data[pos+1];    
        let b = imageData.data[pos+2];
        let a = imageData.data[pos+3];
        
        if(a < startColor.a - tolerence || a > startColor.a + tolerence) {
            return false;
        }

        if(r < startColor.r - tolerence || r > startColor.r + tolerence) {
            return false;
        }
        
        if(g < startColor.g - tolerence || g > startColor.g + tolerence) {
            return false;
        }
        
        if(b < startColor.b - tolerence || b > startColor.b + tolerence) {
            return false;
        }

        return true;
    }

})();