/* global floodFill */
/* global hexToRgb */
/* global webAppCore */
(function() {
    
    // Global Exposure
    window.wacMenuBack = onMenuOptionSelected;

    // Event Linkup
    window.addEventListener("load",startApp,false);
    window.addEventListener("resize",sizeCanvas,false);
    
    let paintArea;
    let paintContext;

    let applicatorArea;
    let applicatorContext;

    let isFlooding = false;
    let hasHadNoHalfPressure = false;

    let selectedPen = 2;

    let stampImage = null;
    let repaintTimeout = null;

    let rgbaColor = null;

    let blotSetter = null;

    let curStampName = null;

    let paintQueue;

    let inToolChangeMode = false;
    let pendingPremiumStamp = null;
    let paintSaveCooldown = null;

    let userTokenData = [];

    let floodTolerence = 50;

    let smallPenVisible = "1";
    let mediumPenVisible = "1";
    let largePenVisible = "1";
    let bucketVisible = "1";
    let stampVisible = "1";
    let eraserVisible = "1";

    let retainedStampImages = {};

    let currentPainting = null;
    let currentPaintingId = null;
    let currentFilename = null;

    const PRIMARY_COLORS_LIGHT = [
        "#EF9A9A",
        "#FFCC80",
        "#FFF59D",
        "#A5D6A7",
        "#90CAF9",
        "#CE93D8",
        "#F48FB1",
        "#BCAAA4",
        "#FFFFFF"
    ];

    const PRIMARY_COLORS = [
        "#F44336",
        "#FF9800",
        "#FFEB3B",
        "#4CAF50",
        "#2196F3",
        "#9C27B0",
        "#E91E63",
        "#795548",
        "#9E9E9E"
    ];

    const PRIMARY_COLORS_DARK = [
        "#B71C1C",
        "#EF6C00",
        "#FBC02D",
        "#1B5E20",
        "#0D47A1",
        "#4A148C",
        "#880E4F",
        "#3E2723",
        "#000000"
    ];

    let STAMPS = [];

    let currentContentHeight = 0;

    let selectedColor = "#F44336";
    let canvasColor = "#ffffff";

    let pointers = {};
    let downPointers = 0;

    let appScale = 1;
    let penSize = 8;

    function startApp() {
        webAppCore.startApp(initApp);
    }

    function initApp() {

        currentPainting = new window.apeApps.fileStructures.fpp.Struct();

        webAppCore.openPackageListing("/stamps/",onStamplistBack);

        webAppCore.setApeAppsStatusListener(onApeAppsStatusChange);
        webAppCore.setOnPremiumListener(onPremiumUpgraded);

        let bucketVisibleDefault = "1";

        if (webAppCore.getInterface() == "android" || (webAppCore.getInterface() == "windows" && window.innerWidth < 500)) {
            bucketVisibleDefault = "0";
        }

        smallPenVisible = webAppCore.getVar("smallPenVisible","1");
        mediumPenVisible = webAppCore.getVar("mediumPenVisible","1");
        largePenVisible = webAppCore.getVar("largePenVisible","1");
        bucketVisible = webAppCore.getVar("bucketVisible",bucketVisibleDefault);
        stampVisible = webAppCore.getVar("stampVisible","1");
        eraserVisible = webAppCore.getVar("eraserVisible","1");

        webAppCore.setTheme("#917c6f");
        
        if(window.devicePixelRatio) {
            appScale = window.devicePixelRatio;
        }

        linkLayouts();
        
        let savedPen = webAppCore.getVar("pen-color","#F44336");
        let savedCanvas = webAppCore.getVar("canvas-color","#ffffff");
        
        setCanvasColor(savedCanvas);
        setColor(savedPen);
        loadPallet();

        webAppCore.handleInput(applicatorArea,handleDown,handleMove,handleUp);
    
        sizeCanvas();
        
        penSize = 8;

        getAccountHolderTokens();

        setTimeout(cleanSlate,100);

        webAppCore.handleBackFunction(null);
        webAppCore.setRelaunchListener(onAppRelaunch);

        webAppCore.checkForLaunchFile(function(has){
            if(has) {
                webAppCore.getLaunchFileName(function(name){
                    if(name) {
                        webAppCore.getLaunchFileContent(function(data){
                            if(data && data.trim().length > 0) {

                                window.apeApps.utilities.compressionHelper.decompress({
                                    data: data,
                                    callback: function(raw) {
                                        if(!raw) {
                                            setTimeout(restoreLastWork,1000);
                                            return;
                                        }
                        
                                        cleanSlate();
                        
                                        currentPainting = window.apeApps.fileStructures.fpp.rebuildFromJSON(raw);
                                        currentFilename = name;
                                        currentPaintingId = "LAUNCH";
                        
                                        doHistoryRepaint();

                                        webAppCore.clearSplash();
                                    }
                                });
                            } else {
                                setTimeout(restoreLastWork,1000);
                            }
                        });
                    } else {
                        setTimeout(restoreLastWork,1000);
                    }
                });
            } else {
                setTimeout(restoreLastWork,1000);
            }
        });

    }

    function onAppRelaunch(type) {
        if(type == "file") {
            webAppCore.checkForLaunchFile(function(has){
                if(has) {
                    webAppCore.getLaunchFileName(function(name){
                        if(name) {
                            webAppCore.getLaunchFileContent(function(data){
                                if(data && data.trim().length > 0) {
                                    currentFilename = name;
 
                                    window.apeApps.utilities.compressionHelper.decompress({
                                        data: data,
                                        callback: function(raw) {
                                            if(!raw) {
                                                return;
                                            }
                            
                                            cleanSlate();
                            
                                            currentPainting = window.apeApps.fileStructures.fpp.rebuildFromJSON(raw);
                                            currentFilename = name;
                                            currentPaintingId = "LAUNCH";
                            
                                            doHistoryRepaint();
    
                                            webAppCore.clearSplash();
                                        }
                                    });
                                }
                            });
                        }
                    });
                }
            });
        }
    }

    function defaultOutColors() {
        setCanvasColor("#ffffff");
        setColor("#F44336");
    }

    function restoreLastWork() {
        let lastWork = webAppCore.getVar("last-paint-hist","0");

        if(lastWork == "0") {
            currentPainting.data = [];
        } else {
            try {
                currentPainting.data = JSON.parse(lastWork);

                if(currentPainting.data.length == 0) {
                    currentPainting.data = [];
                }
            } catch(ex) {
                currentPainting.data = [];
            }
        }

        renderPaintHistory();

        webAppCore.clearSplash();
    }
    
    function gooseUpFlyout() {
        webAppCore.clearMenu();

        let useColor = canvasColor;

        if(useColor == "#ffffff") {
            useColor = null;
        }
        
        webAppCore.addMenuOption("New Painting","wacMenuBack","document-outline",null,"new");

        if(webAppCore.fileDialogsSupported(false)) {
            webAppCore.addMenuOption("Open Painting","wacMenuBack","folder-open-outline",null,"open");
            webAppCore.addMenuOption("Save Painting","wacMenuBack","save-outline",null,"save");
            webAppCore.addMenuOption("Save Painting As","wacMenuBack","winicon.&#xE28F;",null,"saveas");
        }
        

        webAppCore.addMenuOption("Share Painting","wacMenuBack","share-social-outline",null,"share");
        webAppCore.addMenuOption("Print Painting","wacMenuBack","print-outline",null,"print");
        webAppCore.addMenuOption("Undo","wacMenuBack","arrow-undo-outline",null,"undo");
        webAppCore.addMenuOption("Canvas Color","wacMenuBack","color-fill-outline",useColor,"canvas");
        webAppCore.addMenuOption("Tool Settings","wacMenuBack","hammer-outline",null,"settings");
    }



    function onMenuOptionSelected(name,tag) {

        if(tag == "open") {
            webAppCore.openFileDialog(window.apeApps.fileStructures.fpp.EXT,onDialogFileOpened,false);
        }

        if(tag == "undo") {
            if(currentPainting.data.length > 0) {
                currentPainting.data.pop();
                renderPaintHistory();
            }

            savePaintHist();
        }

        if(tag == "settings") {

            if(inToolChangeMode) {
                exitToolVisibilityMode();
            } else {
                var clickWork = "Click";
                
                if(webAppCore.getInterface() == "android" || webAppCore.getInterface == "ios") {
                    clickWork = "Tap";
                }

                enterToolChangeMode();
    
                webAppCore.showAlertWithTitle("Toggle which tools you wish to show or hide.  " + clickWork + " on the paint canvas to save your changes.","Tool Settings");
            }
            
        }

        if(tag == "new") {
            confirmNewImage();
        }
        
        if(tag == "save") {
            saveCurrentPaintint();
        }

        if(tag == "saveas") {
            saveCurrentPaintingAs();
        }

        if(tag == "share") {
            sharePainting();
        }
        
        if(tag == "canvas") {
            canvasColorPicker();
        }
        
        if(tag == "print") {
            let imageData = paintArea.toDataURL();
            let imageTag;

            if(paintArea.width > paintArea.height) {
                imageTag = "<div style=\"position: fixed; top: 0px; left: 0px; height: 100%; width: 100%; -webkit-print-color-adjust:exact; background-color: " + canvasColor + ";\"><img src=\"" + imageData + "\" style=\"position: absolute; top: 50%; left: 50%; transform: translate(-50%, -50%); width:100%\" /></div>";
            } else {
                imageTag = "<div style=\"position: fixed; top: 0px; left: 0px; height: 100%; width: 100%; -webkit-print-color-adjust:exact; background-color: " + canvasColor + ";\"><img src=\"" + imageData + "\" style=\"position: absolute; top: 50%; left: 50%; transform: translate(-50%, -50%); height:100%\" /></div>";                
            }
            
            webAppCore.printHTML(imageTag);
        }
    }

    function populateStamps() {

        let stampArea = document.getElementById("stampArea");
        stampArea.innerHTML = "";

        let i,stampItem;

        for(i = 0; i < userTokenData.length; i++) {

            stampItem = document.createElement("img");
            stampItem.className = "stampItem";
            stampItem.draggable = false;
            stampItem.stampRef = userTokenData[i].url;
            stampItem.stampIdRef = "token-" + userTokenData[i].id;
            stampItem.title = userTokenData[i].name;
            stampItem.onclick = onStampItemTokenClick;
            stampItem.src = userTokenData[i].url;
            stampItem.crossOrigin= "anonymous";
            stampArea.appendChild(stampItem);
        }

        for(i = 0; i < STAMPS.length; i++) {
            stampItem = document.createElement("img");
            stampItem.className = "stampItem";
            stampItem.draggable = false;
            stampItem.stampRef = STAMPS[i];
            stampItem.stampIdRef = "stamp-" + STAMPS[i];
            stampItem.title = STAMPS[i];
            stampItem.onclick = stampItemClick;
            stampItem.src = "stamps/" + STAMPS[i] + ".png";
            stampArea.appendChild(stampItem);
        }   
    }

    function onStampItemTokenClick() {
        document.getElementById("stampArea").style.display = "none";

        pendingPremiumStamp = null;

        clearToolSelection();
        document.getElementById("stamper").className = "tool selected";
        applicatorContext.clearRect(0,0,applicatorArea.width,applicatorArea.height);

        selectedPen = 6;
        webAppCore.logEvent("Tool","Selected","Stamp");

        curStampName = this.stampIdRef;

        stampImage = new Image();
        stampImage.crossOrigin= "anonymous";
        stampImage.src = this.stampRef;
    }

    function stampItemClick() {
        document.getElementById("stampArea").style.display = "none";

        if(!webAppCore.getIsPremium()) {
            pendingPremiumStamp = this.stampRef;
            nagForPremium();
            return;
        }

        webAppCore.logEvent("Stamp","Selected",this.stampRef);
        applicatorContext.clearRect(0,0,applicatorArea.width,applicatorArea.height);

        clearToolSelection();
        document.getElementById("stamper").className = "tool selected";

        curStampName = this.stampIdRef;
        

        selectedPen = 6;
        webAppCore.logEvent("Tool","Selected","Stamp");

        stampImage = new Image();
        stampImage.src = "stamps/" + this.stampRef + ".png";
    }
    
    function linkLayouts() {
        let colorTool = document.getElementById("selectedInk");
        colorTool.style.backgroundColor = selectedColor;
        colorTool.style.borderColor = darkenColor(selectedColor);
        
        let wrapper = document.getElementById("wrapper");
        wrapper.style.backgroundColor = "#ffffff";
        
        colorTool.addEventListener("click",function(e) {
            e.preventDefault();
            pendingPremiumStamp = null;
            togglePicker();
        },false);

        document.getElementById("menuToggle").addEventListener("click",function(e) {
            e.preventDefault();
            webAppCore.toggleMenu();
        },false);

        paintArea = document.getElementById("canvasPaint");
        paintContext = paintArea.getContext("2d");

        applicatorArea = document.getElementById("canvasApplicator");
        applicatorContext = applicatorArea.getContext("2d");
        
        document.getElementById("penSmall").addEventListener("click",function(e) {
            e.preventDefault();

            pendingPremiumStamp = null;
            applicatorContext.clearRect(0,0,applicatorArea.width,applicatorArea.height);

            if(inToolChangeMode) {

                if(smallPenVisible) {
                    smallPenVisible = false;
                    webAppCore.storeVar("smallPenVisible","0");
                } else {
                    smallPenVisible = true;
                    webAppCore.storeVar("smallPenVisible","1");
                }

                enterToolChangeMode();

                return;
            }


            clearToolSelection();
            penSize = 2;
            selectedPen = 2;
            webAppCore.logEvent("Tool","Selected","Small Pen");
            this.className = "tool selected";
            
        },false);
        
        document.getElementById("penMedium").addEventListener("click",function(e) {
            e.preventDefault();

            pendingPremiumStamp = null;
            applicatorContext.clearRect(0,0,applicatorArea.width,applicatorArea.height);

            if(inToolChangeMode) {
                
                if(mediumPenVisible) {
                    mediumPenVisible = false;
                    webAppCore.storeVar("mediumPenVisible","0");
                } else {
                    mediumPenVisible = true;
                    webAppCore.storeVar("mediumPenVisible","1");
                }

                enterToolChangeMode();

                return;
            }

            clearToolSelection();
            penSize = 8;
            selectedPen = 2;
            webAppCore.logEvent("Tool","Selected","Medium Pen");
            this.className = "tool selected";
        },false);
        
        document.getElementById("penLarge").addEventListener("click",function(e) {
            e.preventDefault();

            pendingPremiumStamp = null;
            applicatorContext.clearRect(0,0,applicatorArea.width,applicatorArea.height);

            if(inToolChangeMode) {
                
                if(largePenVisible) {
                    largePenVisible = false;
                    webAppCore.storeVar("largePenVisible","0");
                } else {
                    largePenVisible = true;
                    webAppCore.storeVar("largePenVisible","1");
                }

                enterToolChangeMode();

                return;
            }

            clearToolSelection();
            penSize = 22;
            selectedPen = 2;
            webAppCore.logEvent("Tool","Selected","Large Pen");
            this.className = "tool selected";
        },false);
        
        document.getElementById("paintBucket").addEventListener("click",function(e) {
            e.preventDefault();

            pendingPremiumStamp = null;
            applicatorContext.clearRect(0,0,applicatorArea.width,applicatorArea.height);

            if(inToolChangeMode) {
                
                if(bucketVisible) {
                    bucketVisible = false;
                    webAppCore.storeVar("bucketVisible","0");
                } else {
                    bucketVisible = true;
                    webAppCore.storeVar("bucketVisible","1");
                }

                enterToolChangeMode();

                return;
            }

            clearToolSelection();
            selectedPen = 4;
            webAppCore.logEvent("Tool","Selected","Fill Bucket");
            this.className = "tool selected";
        },false);
        
        document.getElementById("stamper").addEventListener("click",function(e) {
            e.preventDefault();

            pendingPremiumStamp = null;
            applicatorContext.clearRect(0,0,applicatorArea.width,applicatorArea.height);

            if(inToolChangeMode) {
                
                if(stampVisible) {
                    stampVisible = false;
                    webAppCore.storeVar("stampVisible","0");
                } else {
                    stampVisible = true;
                    webAppCore.storeVar("stampVisible","1");
                }

                enterToolChangeMode();

                return;
            }

            if(document.getElementById("stampArea").style.display == "block") {
                document.getElementById("stampArea").style.display = "none";
            } else {
                document.getElementById("stampArea").style.display = "block";
            }

        },false);

        document.getElementById("eraser").addEventListener("click",function(e) {
            e.preventDefault();

            pendingPremiumStamp = null;
            applicatorContext.clearRect(0,0,applicatorArea.width,applicatorArea.height);

            if(inToolChangeMode) {
                
                if(eraserVisible) {
                    eraserVisible = false;
                    webAppCore.storeVar("eraserVisible","0");
                } else {
                    eraserVisible = true;
                    webAppCore.storeVar("eraserVisible","1");
                }

                enterToolChangeMode();

                return;
            }

            clearToolSelection();

            selectedPen = 5;
            webAppCore.logEvent("Tool","Selected","Eraser");

            this.className = "tool selected";
        },false);
  
        setToolVisibility();
    }

    function setToolVisibility() {
        if(smallPenVisible == "1") {
            document.getElementById("penSmall").style.display = null;
        } else {
            document.getElementById("penSmall").style.display = "none";
        }

        if(mediumPenVisible == "1") {
            document.getElementById("penMedium").style.display = null;
        } else {
            document.getElementById("penMedium").style.display = "none";
        }

        if(largePenVisible == "1") {
            document.getElementById("penLarge").style.display = null;
        } else {
            document.getElementById("penLarge").style.display = "none";
        }

        if(bucketVisible == "1") {
            document.getElementById("paintBucket").style.display = null;
        } else {
            document.getElementById("paintBucket").style.display = "none";
        }

        if(stampVisible == "1") {
            document.getElementById("stamper").style.display = null;
        } else {
            document.getElementById("stamper").style.display = "none";
        }

        if(eraserVisible == "1") {
            document.getElementById("eraser").style.display = null;
        } else {
            document.getElementById("eraser").style.display = "none";
        }
    }

    function exitToolVisibilityMode() {
        inToolChangeMode = false;

        document.getElementById("penSmall").style.backgroundColor = null;
        document.getElementById("penMedium").style.backgroundColor = null;
        document.getElementById("penLarge").style.backgroundColor = null;
        document.getElementById("paintBucket").style.backgroundColor = null;
        document.getElementById("stamper").style.backgroundColor = null;
        document.getElementById("eraser").style.backgroundColor = null;

        setToolVisibility();
    }
    
    function clearToolSelection() {
        document.getElementById("stampArea").style.display = "none";
        document.getElementById("penSmall").className = "tool";
        document.getElementById("penMedium").className = "tool";
        document.getElementById("penLarge").className = "tool";
        document.getElementById("paintBucket").className = "tool";
        document.getElementById("stamper").className = "tool";
        document.getElementById("eraser").className = "tool";
        stampImage = null;
    }

    function enterToolChangeMode() {
        inToolChangeMode = true;

        document.getElementById("penSmall").style.display = null;
        document.getElementById("penMedium").style.display = null;
        document.getElementById("penLarge").style.display = null;
        document.getElementById("paintBucket").style.display = null;
        document.getElementById("stamper").style.display = null;
        document.getElementById("eraser").style.display = null;

        if(smallPenVisible == "1") {
            document.getElementById("penSmall").style.backgroundColor = "#4CAF50";
        } else {
            document.getElementById("penSmall").style.backgroundColor = "#F44336";
        }

        if(mediumPenVisible == "1") {
            document.getElementById("penMedium").style.backgroundColor = "#4CAF50";
        } else {
            document.getElementById("penMedium").style.backgroundColor = "#F44336";
        }

        if(largePenVisible == "1") {
            document.getElementById("penLarge").style.backgroundColor = "#4CAF50";
        } else {
            document.getElementById("penLarge").style.backgroundColor = "#F44336";
        }

        if(bucketVisible == "1") {
            document.getElementById("paintBucket").style.backgroundColor = "#4CAF50";
        } else {
            document.getElementById("paintBucket").style.backgroundColor = "#F44336";
        }

        if(stampVisible == "1") {
            document.getElementById("stamper").style.backgroundColor = "#4CAF50";
        } else {
            document.getElementById("stamper").style.backgroundColor = "#F44336";
        }

        if(eraserVisible == "1") {
            document.getElementById("eraser").style.backgroundColor = "#4CAF50";
        } else {
            document.getElementById("eraser").style.backgroundColor = "#F44336";
        }
    }

    function confirmNewImage() {

        webAppCore.showDialog({
            title: webAppCore.getString("clear_painting"),
            message: webAppCore.getString("clear_painting_msg"),
            op1: {
                text: webAppCore.getString("gen_yes"),
                func: function(){
                    defaultOutColors();
                    cleanSlate();
                },
                color: "#ff0000"
            },
            op2: {
                text: webAppCore.getString("gen_no"),
                func: placeholder
            }
        });

    }
    
    function placeholder() {}

    function savePaintHist() {

        if(paintSaveCooldown) {
            clearTimeout(paintSaveCooldown);
            paintSaveCooldown = null;
        }


        paintSaveCooldown = setTimeout(executePaintHistSave,1500);
        
    }

    function executePaintHistSave() {

        if(!currentPainting.data) {
            currentPainting.data = [];
        }

        var paintString = JSON.stringify(currentPainting.data);
        webAppCore.storeVar("last-paint-hist",paintString);
    }
    
    function cleanSlate() {
        currentFilename = null;
        currentPaintingId = null;
        currentPainting = new window.apeApps.fileStructures.fpp.Struct();
        savePaintHist();
        sizeCanvas();
    }
    
    function sizeCanvas() {

        if(repaintTimeout) {
            clearTimeout(repaintTimeout);
        }

        repaintTimeout = null;

        let toolBox = document.getElementById("toolBox");
        currentContentHeight = toolBox.offsetHeight; 
        
        let displayWidth = window.innerWidth;
        let displayHeight = (window.innerHeight - currentContentHeight);

        let scaleWidth = displayWidth * appScale;
        let scaleHeight = displayHeight * appScale;

        let wrapper = document.getElementById("wrapper");
        wrapper.style.top = currentContentHeight + "px";
        wrapper.style.height = displayHeight + "px";
        wrapper.style.width = displayWidth + "px";
    
        if(paintArea) {
            paintArea.width = scaleWidth;
            paintArea.height = scaleHeight;
            paintArea.style.width = displayWidth + "px";
            paintArea.style.height = displayHeight + "px";

            paintContext.clearRect(0,0,paintArea.width,paintArea.height);
        }
        
        if(applicatorArea) {
            applicatorArea.width = displayWidth;
            applicatorArea.height = displayHeight;
            applicatorArea.style.width = displayWidth + "px";
            applicatorArea.style.height = displayHeight + "px";
        
            
            applicatorContext.clearRect(0,0,applicatorArea.width,applicatorArea.height);

        }
    
        repaintTimeout = setTimeout(doHistoryRepaint,200);
    }

    function doHistoryRepaint() {

        for(let i = 0; i < currentPainting.data.length; i++) {
            paintPointers(currentPainting.data[i]);
        }

        repaintTimeout = null;
    }

    function PaintCoord(x,y,pressure,id) {
        this.x = x * appScale;
        this.y = y * appScale;
        this.pressure = pressure;
        this.id = id;
    }

    

    function handleDown(target, pointerId, x, y,type,pressure) {

        if(inToolChangeMode) {
            exitToolVisibilityMode();
            return;
        }
        
        if(isFlooding) {
            return;
        }

        if(!paintQueue) {
            paintQueue = [];
        }

        if(selectedPen == 4) {
            x = x * appScale;
            y = (y - currentContentHeight) * appScale;
            isFlooding = true;
            floodFill(paintContext,x,y,selectedColor,floodTolerence);
            isFlooding = false;

            // add to undo queue 
            currentPainting.data.push({
                "0": {
                    "p": [{
                        "x": x,
                        "y": y
                    }],
                    "ap": 1,
                    "tp": 1,
                    "t": "fill",
                    "c": selectedColor
                }
            });

            savePaintHist();

            return;
        }

        // for the applicator canvas 
        if(!pointers[pointerId]) {
            pointers[pointerId] = {};
        }

        if(selectedPen == 5) {
            pointers[pointerId].down = true;
            handleMove(target, pointerId, x, y,type,pressure);
            
            return;
        }

        if(selectedPen == 6) {
            if(stampImage != null) {
                paintContext.drawImage(stampImage,(x - 24) * appScale,((y - 24) * appScale) - currentContentHeight,48 * appScale,48 * appScale);

                // add to undo queue 
                currentPainting.data.push({
                    "0": {
                        "p": [{
                            "x": (x - 24) * appScale,
                            "y": ((y - 24) * appScale) - currentContentHeight
                        }],
                        "ap": 1,
                        "tp": 1,
                        "t": "stamp",
                        "c": curStampName
                    }
                });

                savePaintHist();
            }

            return;
        }

        y -= currentContentHeight;

        let usePressure = 1;

        if(pressure && pressure != 0.5) {
            usePressure = pressure;
        }

        if(type == "touch") {
            usePressure *= 2;

            if(usePressure > 1) {
                usePressure = 1;
            }
        }

        paintQueue.push(new PaintCoord(x,y,usePressure,pointerId));

        
    
        pointers[pointerId].down = true;
        pointers[pointerId].moved = false;
        pointers[pointerId].x = x;
        pointers[pointerId].y = y;
        pointers[pointerId].oldx = x;
        pointers[pointerId].oldy = y;

    }
        
    function handleMove(target, pointerId, x, y,type,pressure) {

        if(isFlooding || selectedPen == 4) {
            return;
        }

        if(selectedPen == 6 && stampImage != null && type == "mouse") {
            applicatorContext.clearRect(0,0,applicatorArea.width,applicatorArea.height);
            applicatorContext.save();
            applicatorContext.globalAlpha = 0.4;
            applicatorContext.drawImage(stampImage,(x - 24) * appScale,((y - 24) * appScale) - currentContentHeight,48 * appScale,48 * appScale);
            applicatorContext.restore();
            return;
        }

        if(!pointers[pointerId]) {
            pointers[pointerId] = {};
        }

        if(!pointers[pointerId].down) {
            return;
        }

        let usePressure = 1;

        if(pressure) {
            if(pressure > 0 && pressure < 1 && pressure != 0.5) {
                hasHadNoHalfPressure = true;
            }

            if(pressure != 0.5 || (pressure == 0.5 && hasHadNoHalfPressure)) {
                usePressure = pressure;
            }

            
        }

        if(type == "touch") {
            usePressure *= 2;

            if(usePressure > 1) {
                usePressure = 1;
            }
        }

        y -= currentContentHeight;

        if(pointers[pointerId].x != x || pointers[pointerId].y != y) {
            pointers[pointerId].moved = true;
        }
        
        pointers[pointerId].x = x;
        pointers[pointerId].y = y;

        if(selectedPen == 5) {

            let rad = Math.round(30 * appScale);
            let ux = Math.round(x * appScale);
            let uy = Math.round(y * appScale);

            clearCircle(paintContext,ux,uy,rad);

            paintQueue.push(new PaintCoord(x,y,rad,pointerId));
            
            return;
        }

        paintQueue.push(new PaintCoord(x,y,usePressure,pointerId));

        setRGBColor(usePressure);

        if(selectedPen == 4) {
            pointers[pointerId].oldx = pointers[pointerId].x;
            pointers[pointerId].oldy = pointers[pointerId].y;
            return;
        }
    
        applicatorContext.beginPath();
        applicatorContext.moveTo(pointers[pointerId].oldx, pointers[pointerId].oldy);
        applicatorContext.lineTo(pointers[pointerId].x, pointers[pointerId].y);
    
        applicatorContext.lineWidth = penSize;
    
        applicatorContext.strokeStyle = rgbaColor;

        applicatorContext.lineCap = "round";

        if(penSize < (8 * appScale)) {
            if(usePressure < 0.75) {
                applicatorContext.lineCap = "butt";
            }
        }

        applicatorContext.stroke();
    
        pointers[pointerId].oldx = pointers[pointerId].x;
        pointers[pointerId].oldy = pointers[pointerId].y;

    }
        
    function handleUp(target, pointerId) {
        
        downPointers--;
        
        if(downPointers <= 0) {
            downPointers = 0;
        }


        if(!pointers[pointerId]) {
            return;
        }


        if(selectedPen != 5 && pointers[pointerId].down && !pointers[pointerId].moved) {

            let pw = (penSize / 2);

            applicatorContext.beginPath();
            applicatorContext.arc(pointers[pointerId].x,pointers[pointerId].y, pw, 0, 2 * Math.PI, false);
            applicatorContext.fillStyle = rgbaColor;
            applicatorContext.fill();
        }

        pointers[pointerId].down = false;

        engagePaintQueue();
        
    }

    function engagePaintQueue() {

        if(!paintQueue || paintQueue.length == 0) {
            return;
        }

        let pointerQueues = {};

        let drawTip = penSize * appScale;

        let totaloperations = 0;

        while(paintQueue.length > 0) {
            let pq = paintQueue.shift();

            let useCol = selectedColor;
            let useTip = drawTip;

            if(selectedPen == 5) {
                useTip = "erase";
                useCol = pq.pressure;
            }

            if(!pointerQueues[pq.id]) {
                pointerQueues[pq.id] = {
                    "p": [],
                    "ap": 0,
                    "tp": 0,
                    "t": useTip,
                    "c": useCol
                };
            }

            pointerQueues[pq.id].p.push({
                "x": pq.x,
                "y": pq.y
            });

            pointerQueues[pq.id].tp += pq.pressure;

            totaloperations++;
        }

        if(totaloperations == 0) {
            return;
        }

        currentPainting.data.push(pointerQueues);
        savePaintHist();

        // engage 
        paintPointers(pointerQueues);
        

        applicatorContext.clearRect(0,0,applicatorArea.width,applicatorArea.height);

    }

    function renderPaintHistory() {
        sizeCanvas();
        doHistoryRepaint();
    }

    function paintPointers(pointerQueues) {
        
        let i,useColor;

        for(let pointer in pointerQueues) {
            pointerQueues[pointer].ap = pointerQueues[pointer].tp / pointerQueues[pointer].p.length;

            if(pointerQueues[pointer].t == "fill" || pointerQueues[pointer].t == "stamp" || pointerQueues[pointer].t == "erase") {
                useColor = null;
            } else {
                useColor = getrgbaColor(pointerQueues[pointer].c,pointerQueues[pointer].ap);
            }

            paintContext.beginPath();

            if(pointerQueues[pointer].p.length > 1) {

                if(pointerQueues[pointer].t == "erase") {
                    for(i = 1; i < pointerQueues[pointer].p.length - 2; i++) {
                        clearCircle(paintContext,pointerQueues[pointer].p[i].x,pointerQueues[pointer].p[i].y,pointerQueues[pointer].c);
                    }
                    continue;
                }

                paintContext.lineWidth = pointerQueues[pointer].t;
                paintContext.strokeStyle = useColor;
                paintContext.lineCap = "round";

                paintContext.moveTo(pointerQueues[pointer].p[0].x,pointerQueues[pointer].p[0].y);

                for(i = 1; i < pointerQueues[pointer].p.length - 2; i++) {
                    let c = (pointerQueues[pointer].p[i].x + pointerQueues[pointer].p[i + 1].x) / 2;
                    let d = (pointerQueues[pointer].p[i].y + pointerQueues[pointer].p[i + 1].y) / 2;

                    paintContext.quadraticCurveTo(pointerQueues[pointer].p[i].x, pointerQueues[pointer].p[i].y, c, d);
                }

                if(pointerQueues.length > 2) {
                    paintContext.quadraticCurveTo(
                        pointerQueues[pointer].p[i].x,
                        pointerQueues[pointer].p[i].y,
                        pointerQueues[pointer].p[i + 1].x,
                        pointerQueues[pointer].p[i + 1].y
                    );
                }

                paintContext.stroke();
            } else {

                if(pointerQueues[pointer].t == "fill") {
                    floodFill(paintContext,pointerQueues[pointer].p[0].x,pointerQueues[pointer].p[0].y,pointerQueues[pointer].c,floodTolerence);
                    continue;
                }

                if(pointerQueues[pointer].t == "stamp") {
                    doStampRepaint(pointerQueues[pointer].c,pointerQueues[pointer].p[0].x,pointerQueues[pointer].p[0].y);

                    continue;
                }

                if(pointerQueues[pointer].t == "erase") {
                    clearCircle(paintContext,pointerQueues[pointer].p[0].x,pointerQueues[pointer].p[0].y,pointerQueues[pointer].c);
                    continue;
                }

                let pw = (pointerQueues[pointer].t / 2);

                paintContext.beginPath();
                paintContext.arc(pointerQueues[pointer].p[0].x,pointerQueues[pointer].p[0].y, pw, 0, 2 * Math.PI, false);
                paintContext.fillStyle = useColor;
                paintContext.fill();
            }
        }
    }

    function getStampImageByRefName(refname) {

        if(retainedStampImages[refname]) {

            if(retainedStampImages[refname].loading) {
                return null;
            }

            return retainedStampImages[refname].img;
        }

        let allStamps = document.querySelectorAll(".stampItem");

        for(let i = 0; i < allStamps.length; i++) {
            if(allStamps[i].stampIdRef == refname) {
                let retImg = new Image();

                retainedStampImages[refname] = {
                    loading: true,
                    img: retImg
                };

                retImg.onload = function(){
                    retainedStampImages[refname].loading = false;
                };

                if(refname.indexOf("token-") == 0) {
                    retImg.crossOrigin= "anonymous";
                    retImg.src = allStamps[i].stampRef;
                } else {
                    retImg.src = "stamps/" + allStamps[i].stampRef + ".png";
                }

                

                break;

            }
        }

        return null;
    }

    function loadPallet() {
        let i,blot;

        for(i = 0; i < PRIMARY_COLORS_LIGHT.length; i++) {
            blot = document.createElement("div");
            blot.className = "inkBox";
            blot.style.backgroundColor = PRIMARY_COLORS_LIGHT[i];
            blot.style.borderColor = darkenColor(PRIMARY_COLORS_LIGHT[i]);
            blot.colorReference = PRIMARY_COLORS_LIGHT[i];

            blot.addEventListener("click",onBlotClick);

            document.getElementById("extendedAreaLight").appendChild(blot);
        }

        for(i = 0; i < PRIMARY_COLORS.length; i++) {
            blot = document.createElement("div");
            blot.className = "inkBox";
            blot.style.backgroundColor = PRIMARY_COLORS[i];
            blot.style.borderColor = darkenColor(PRIMARY_COLORS[i]);
            blot.colorReference = PRIMARY_COLORS[i];

            blot.addEventListener("click",onBlotClick);

            document.getElementById("extendedArea").appendChild(blot);
        }

        for(i = 0; i < PRIMARY_COLORS_DARK.length; i++) {
            blot = document.createElement("div");
            blot.className = "inkBox";
            blot.style.backgroundColor = PRIMARY_COLORS_DARK[i];
            blot.style.borderColor = darkenColor(PRIMARY_COLORS_DARK[i]);
            blot.colorReference = PRIMARY_COLORS_DARK[i];

            blot.addEventListener("click",onBlotClick);

            document.getElementById("extendedAreaDark").appendChild(blot);
        }

        for(i = 0; i < PRIMARY_COLORS.length; i++) {
            blot = document.createElement("div");
            blot.className = "inkBox";
            blot.id = "blotItem" + i;
            blot.idReference = i;

            let theColor = webAppCore.getVar("custom-" + i,"0");

            if(theColor == "0") {
                blot.style.backgroundColor = "#ffffff";
                blot.style.borderColor = darkenColor("#ffffff");
            } else {
                blot.style.backgroundColor = theColor;
                blot.style.borderColor = darkenColor(theColor);
            }

            blot.colorReference = theColor;
            blot.addEventListener("click",onCustomBlotClick);
            document.getElementById("customArea").appendChild(blot);
        }
    }

    function onBlotClick() {
        blotSetter = null;
        setColor(this.colorReference);
    }

    function onCustomBlotClick() {
        if(selectedColor == this.colorReference || this.colorReference == "0") {
            blotSetter = this.idReference;
            webAppCore.pickColor(setColor,selectedColor);
        } else {
            blotSetter = null;
            setColor(this.colorReference);
        }
    }
    
    function setColor(color) {

        if(!color) {
            return;
        }
        
        selectedColor = color;
        setRGBColor(1);
        
        let colorTool = document.getElementById("selectedInk");
        colorTool.style.backgroundColor = color;
        colorTool.style.borderColor = darkenColor(color);
        
        webAppCore.logEvent("Color","Selected",selectedColor);
        
        webAppCore.storeVar("pen-color",color);

        if(blotSetter != null) {
            var blotItem = document.getElementById("blotItem" + blotSetter);
            blotItem.style.backgroundColor = color;
            blotItem.style.borderColor = darkenColor(color);
            blotItem.colorReference = color;
            webAppCore.storeVar("custom-" + blotSetter,color);
        }

        document.getElementById("extendedArea").style.display = "none";
        document.getElementById("extendedAreaLight").style.display = "none";
        document.getElementById("extendedAreaDark").style.display = "none";
        document.getElementById("customArea").style.display = "none";

        blotSetter = null;
    }

    function getrgbaColor(color,opacity) {

        let hex = hexToRgb(color);

        if(!hex) {
            return "rgba(0,0,0," + opacity + ")";
        }

        let r = hex.r;
        let g = hex.g;
        let b = hex.b;

        return "rgba(" + r + "," + g + "," + b + "," + opacity + ")";
    }
    
    function setRGBColor(opacity) {

        if(!selectedColor) {
            return;
        }

        rgbaColor = getrgbaColor(selectedColor,opacity);
    }

    function sharePainting() {
        let wrapper = document.getElementById("wrapper");
        webAppCore.sendImage(wrapper);
    }
    
    function togglePicker() {

        if(document.getElementById("extendedArea").style.display == "block") {
            document.getElementById("extendedArea").style.display = "none";
            document.getElementById("extendedAreaLight").style.display = "none";
            document.getElementById("extendedAreaDark").style.display = "none";
            document.getElementById("customArea").style.display = "none";
        } else {
            document.getElementById("extendedArea").style.display = "block";
            document.getElementById("extendedAreaLight").style.display = "block";
            document.getElementById("extendedAreaDark").style.display = "block";
            document.getElementById("customArea").style.display = "block";
        }

        
    }
    
    function canvasColorPicker() {
        webAppCore.pickColor(setCanvasColor,canvasColor);
    }
    
    function setCanvasColor(color) {

        if(!color) {
            return;
        }
        
        canvasColor = color;
        
        var wrapper = document.getElementById("wrapper");
        wrapper.style.backgroundColor = color;

        webAppCore.logEvent("Canvas Color","Selected",canvasColor);
        
        gooseUpFlyout();
        
        webAppCore.storeVar("canvas-color",color);
    }

    function darkenColor(color) {

        if(!color || color.trim().length != "7") {
            return("rgba(200,0,0,1)");
        }

        let r = hexToRgb(color).r;
        let g = hexToRgb(color).g;
        let b = hexToRgb(color).b;
        
        if(r > 40) {
            r -= 40;
        } else {
            r = 0;
        }
        
        if(g > 40) {
            g -= 40;
        } else {
            g = 0;
        }
        
        if(b > 40) {
            b -= 40;
        } else {
            b = 0;
        }
        
        return("rgba(" + r + "," + g + "," + b + ",1)");
    }
    
    function saveImage() {}

    function nagForPremium() {
        let msg = "Remove advertising and unlock " + STAMPS.length + " (and counting) fun stickers by upgrading to the full version of Finger Paint!  It's cheap and it supports continued development!";

        let dialogOptions = {
            "title": "Go Premium",
            "message": msg,
            "op1": {
                "text": "Upgrade",
                "func": function() {
                    webAppCore.presentPremium();
                }
            },
            "op2": {
                "text": "No way!",
                "func": function() {
                    pendingPremiumStamp = null;
                }
            }
        };

        webAppCore.showDialog(dialogOptions);
    }

    function onPremiumUpgraded() {

        if(pendingPremiumStamp != null) {
            webAppCore.logEvent("Stamp","Purchased",pendingPremiumStamp);
        }

        pendingPremiumStamp = null;
    }

    function getAccountHolderTokens() {
        userTokenData = [];

        if(webAppCore.isApeAppsLoggedIn()) {
            let tokenSize = 48 * appScale;
            webAppCore.getTokens(webAppCore.getApeAppsProfile().un,tokenSize,"37",onUserTokensBack);
        }
    }

    function onUserTokensBack(result) {

        if(!result) {
            return;
        }

        for(let i = 0; i < result.length; i++) {
            userTokenData.push({
                "name": result[i].name,
                "url": result[i].url,
                "id": result[i].id,
                "isMine": true
            });
        }

        populateStamps();
    }

    function onApeAppsStatusChange(status) {
        userTokenData = [];

        if(status) {
            getAccountHolderTokens();
        } else {
            populateStamps();
        }
    }

    function clearCircle(context,x,y,radius) {
        context.save();
        context.beginPath();
        context.arc(x, y, radius, 0, 2 * Math.PI, false);
        context.clip();
        context.clearRect(x - radius - 1, y - radius - 1, radius * 2 + 2, radius * 2 + 2);
        context.restore();
    }

    function onStamplistBack(list) {
        if(!list) {
            return;
        }

        retainedStampImages = {};
        STAMPS = [];

        for(let i = 0; i < list.length; i++) {
            let file = list[i];

            if(file.toLowerCase().indexOf(".png") == -1) {
                continue;
            }

            STAMPS.push(file.replace(".png",""));
        }

        STAMPS.sort();

        populateStamps();
        renderPaintHistory();
    }

    function doStampRepaint(stamp,x,y) {
        let stmpImgRaw = getStampImageByRefName(stamp);

        if(stmpImgRaw == null) {
            setTimeout(function(){
                doStampRepaint(stamp,x,y);
            },150);

            return;
        }

        paintContext.drawImage(stmpImgRaw,x,y,48 * appScale,48 * appScale);
    }

    function onDialogFileOpened(name,content,fileid) {
        window.apeApps.utilities.compressionHelper.decompress({
            data: content,
            callback: function(raw) {
                if(!raw) {
                    return;
                }

                cleanSlate();

                currentPainting = window.apeApps.fileStructures.fpp.rebuildFromJSON(raw);
                currentFilename = name;
                currentPaintingId = fileid;

                doHistoryRepaint();
            }
        });
    }

    function saveCurrentPaintint() {
        if(!currentFilename || !currentPaintingId) {
            saveCurrentPaintingAs();
            return;
        }

        buildSaveDataString(function(saveData){
            if(!saveData) {
                return;
            }

            if(currentPaintingId == "LAUNCH") {
                webAppCore.saveLaunchFile(saveData);
                webAppCore.showToast("Painting Saved");
            } else {
                webAppCore.filesystemFileSave(saveData,currentPaintingId,onPaintingSaved);
            }

            
        });
    }

    function saveCurrentPaintingAs() {
        buildSaveDataString(function(saveData){
            if(!saveData) {
                return;
            }

            webAppCore.exportFile("painting." + window.apeApps.fileStructures.fpp.EXT,saveData,false,onPaintingSaved);
        });
    }

    function buildSaveDataString(callback) {

        let un = null;

        if(webAppCore.isApeAppsLoggedIn()) {
            let profile = webAppCore.getApeAppsProfile();

            if(profile) {
                un = profile.un;
            }
        }

        currentPainting.lastSave = new Date().getTime();

        if(currentPainting.authors.indexOf(un) == -1) {
            currentPainting.authors.push(un);
        }

        currentPainting.lastAuthor = un;

        if(!currentPainting.createdApp) {
            currentPainting.createdApp = "Finger Paint";
            currentPainting.createdAppVer = webAppCore.getAppVersion();
            currentPainting.firstAuthor = un;
        }

        currentPainting.savedApp = "Finger Paint";
        currentPainting.savedAppVer = webAppCore.getAppVersion();

        let str = JSON.stringify(currentPainting);

        window.apeApps.utilities.compressionHelper.compress({
            data: str,
            callback: function(comp) {
                if(!comp) {
                    callback(null);
                    return;
                }

                callback(comp);
            }
        });
    }

    function onPaintingSaved(fileId,fileName) {
        if(!fileId || !fileName) {
            return;
        }

        currentFilename = fileName;
        currentPaintingId = fileId;

        webAppCore.showToast("Painting Saved");
    }
})();
